#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions of the Report class used by MLonMCU sessions and runs."""
from pathlib import Path
import pandas as pd


pd.set_option("display.max_columns", None)
pd.set_option("display.max_rows", None)
pd.set_option("display.width", 0)

SUPPORTED_FMTS = ["csv", "xlsx"]


class Report:
    """Report class wrapped around multiple pandas dataframes."""

    def __init__(self):
        self.pre_df = pd.DataFrame()
        self.main_df = pd.DataFrame()
        self.post_df = pd.DataFrame()

    @property
    def df(self):
        """Combine the three internal dataframes to a large one and return in."""
        # TODO: handle this properly by either adding NAN or use a single set(pre=, post=, main=) method
        return pd.concat([self.pre_df, self.main_df, self.post_df], axis=1)

    def export(self, path):
        """Export the report to  a file.

        Arguments
        ---------
        path : str
            Destination path.

        """
        ext = Path(path).suffix[1:]
        assert ext in SUPPORTED_FMTS, f"Unsupported report format: {ext}"
        parent = Path(path).parent
        if not parent.is_dir():
            parent.mkdir()
        if ext == "csv":
            self.df.to_csv(path, index=False)
        elif ext in ["xlsx", "xls"]:
            self.df.to_excel(path, index=False)
        else:
            raise RuntimeError()

    # def append(self, *args, **kwargs):
    #     self.df = self.df.append(*args, **kwargs, ignore_index=True)

    def set_pre(self, data):
        """Setter for the left third of the dataframe."""
        self.pre_df = pd.DataFrame.from_records(data).reset_index(drop=True)

    def set_post(self, data):
        """Setter for the right third of the dataframe."""
        self.post_df = pd.DataFrame.from_records(data).reset_index(drop=True)

    def set_main(self, data):
        """Setter for the center part of the dataframe."""
        self.main_df = pd.DataFrame.from_records(data).reset_index(drop=True)

    def set(self, pre=None, main=None, post=None):
        """Setter for the dataframe."""
        size = len(pre)
        self.set_pre(pre if pre is not None else {})
        if len(main) != size:
            assert len(main) == 0
            # self.set_main(pd.Series())
        else:
            self.set_main(main if main is not None else {})
        self.set_post(post if post is not None else {})

    def add(self, reports):
        """Helper function to append a line to an existing report."""
        if not isinstance(reports, list):
            reports = [reports]
        for report in reports:
            self.pre_df = pd.concat([self.pre_df, report.pre_df], axis=0).reset_index(drop=True)
            self.main_df = pd.concat([self.main_df, report.main_df], axis=0).reset_index(drop=True)
            self.post_df = pd.concat([self.post_df, report.post_df], axis=0).reset_index(drop=True)



#!/usr/bin/env python
#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for `mlonmcu` package."""


# import unittest
import pytest

# import os
# import tempfile
import configparser
from pathlib import Path

# from pytest_console_scripts import ScriptRunner
# import xdg
# import mock

from mlonmcu.version import __version__
from mlonmcu.cli.main import main


# @pytest.fixture()
# def fake_config_home(tmp_path):
#     config_home = tmp_path / "cfg"
#     config_home.mkdir()
#     mlonmcu_config_home = config_home / "mlonmcu"
#     mlonmcu_config_home.mkdir()
#     patcher = mock.patch.dict(os.environ, {"XDG_CONFIG_HOME": str(config_home)})
#     patcher.start()
#     yield mlonmcu_config_home
#     patcher.stop()
#
#
# @pytest.fixture()
# def fake_environment_directory(tmp_path):
#     cwd = tmp_path / "home"
#     cwd.mkdir()
#     yield cwd
#
#
# @pytest.fixture()
# def fake_working_directory(tmp_path):
#     cwd = tmp_path / "cwd"
#     cwd.mkdir()
#     yield cwd


@pytest.mark.script_launch_mode("subprocess")
def test_version(capsys):
    with pytest.raises(SystemExit) as pytest_exit:
        main(["--version"])
    out, err = capsys.readouterr()
    assert pytest_exit.value.code == 0
    assert out == f"mlonmcu {__version__}\n"
    assert err == ""


def _write_environments_file(path, data):
    config = configparser.ConfigParser()
    for key in data:
        config[key] = data[key]
    with open(path, "w") as env_file:
        config.write(env_file)
    print("_write_environments_file", path, data)


def _create_empty_environments_file(path):
    _write_environments_file(path, {})
    # with open(path, "w") as env_file:
    #     env_file.write("")


def _create_valid_environments_file(path):
    _write_environments_file(
        path,
        {
            "default": {"path": "/x/y/z", "created": "20211223T08000"},
            "custom": {"path": "/a/b/c"},
        },
    )
    # with open(path, "w") as env_file:
    #     env_file.write("""[default]\npath=\"/x/y/z\"\ncreated=\"20211223T08000\"\n\n[custom]\npath=\"/a/b/c\"""")


def _create_complex_environments_file(path):
    _write_environments_file(
        path,
        {
            "default": {"path": "/x/y/z", "created": "20211223T08000"},
            "custom": {"path": "/a/b/c"},
        },
    )
    # with open(path, "w") as env_file:
    #     env_file.write("""[default]\npath=\"/x/y/z\"\ncreated=\"20211223T08000\"\n\n[custom]\npath=\"/a/b/c\"""")


def _create_invalid_environments_file(path):
    _write_environments_file(path, {"default": {}})
    # with open(path, "w") as env_file:
    #     env_file.write("[default]")


def _count_envs(text):
    count = 0
    for line in text.split("\n"):
        if "    - " in line:
            count = count + 1
    return count


@pytest.mark.script_launch_mode("subprocess")
def test_env_empty(fake_config_home: Path, capsys):
    _create_empty_environments_file(fake_config_home / "environments.ini")
    ret = main(["env"])
    out, err = capsys.readouterr()
    assert ret == 0
    count = _count_envs(out)
    assert count == 0


# @pytest.mark.script_launch_mode("subprocess")
def test_env_valid(fake_config_home: Path, capsys):
    _create_valid_environments_file(fake_config_home / "environments.ini")
    # ret = script_runner.run("mlonmcu", "env")
    ret = main(["env"])
    out, err = capsys.readouterr()
    assert ret == 0
    count = _count_envs(out)
    assert count > 0


# @pytest.mark.script_launch_mode('subprocess')
# def test_env_file(script_runner: ScriptRunner, fake_config_home: Path):
#     ret = script_runner.run('mlonmcu', 'env', '--file')
#     assert ret.success
#     assert dirname in ret.stdout
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_add(script_runner: ScriptRunner, fake_config_home: Path):
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'add', 'new', '/foo/bar')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_add_invalid(script_runner: ScriptRunner, fake_config_home: Path):  # already exists
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'add', 'default', '/foo/bar')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_delete(script_runner: ScriptRunner, fake_config_home: Path):
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'delete', 'custom')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_delete_invalid(script_runner: ScriptRunner, fake_config_home: Path):  # does not exits
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'delete', 'custom2')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_delete_invalid(script_runner: ScriptRunner, fake_config_home: Path):  # does not exits
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'delete', 'custom2')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_update(script_runner: ScriptRunner, fake_config_home: Path):
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'update', 'custom', '/foo/bar')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_update_invalid(script_runner: ScriptRunner, fake_config_home: Path):  # does not exist
#     _create_valid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env', 'update', 'custom2', '/foo/bar')
#     assert ret.success
#     # TODO
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_invalid2(script_runner: ScriptRunner, fake_config_home: Path):
#     _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'env')
#     assert not ret.success

# class MockedEnvironmentDirectory:
#     def __init__(self):
#         self.tempdir = None
#         self.env_dir = None
#         self.env_file_path = None
#     def __enter__(self):
#         self.tempdir = tempfile.TemporaryDirectory()
#         self.env_dir = Path(self.tempdir.name) / "mlonmcu"
#         os.mkdir(self.env_dir)
#         self.env_file_path = self.env_dir / "environments.ini"
#         return self
#     def __exit__(self, type, value, traceback):
#         self.tempdir.cleanup()
#         self.env_dir = None
#         self.env_file_path = None
#
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_env_invalid_compact(script_runner):
#     with MockedEnvironmentDirectory() as mocked:
#         _create_invalid_environments_file(mocked.env_file_path)
#         ret = script_runner.run('mlonmcu', 'env')
#         assert not ret.success

# TODO:
# test_cleanup_all
# test_cleanup_deps
# test_cleanup_results
# test_cleanup_sessions
# test_cleanup_session

# @pytest.mark.script_launch_mode('subprocess')
# def test_cleanup_all(script_runner):
#     with tempfile.TemporaryDirectory() as dirname:
#         env_dir = Path(dirname) / "mlonmcu"
#         os.mkdir(env_dir)
#         env_file_path = env_dir / "environments.ini"
#         _create_valid_environments_file(env_file_path)
#         with mock.patch.dict(os.environ, {"XDG_CONFIG_HOME": dirname}):
#             ret = script_runner.run('mlonmcu', 'env')
#             assert ret.success
#             count = _count_envs(ret.stdout)
#             assert count > 0

# @pytest.mark.script_launch_mode('subprocess')
# def test_init_local(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init', str(fake_environment_directory))
#     # ensure initialized
#     # ensure not registered
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_local_empty(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init', str(fake_environment_directory))
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_local_nonempty(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init', str(fake_environment_directory))
#     assert not ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_local_named(script_runner: ScriptRunner, fake_config_home: Path, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init', str(fake_environment_directory), '--name', 'local')
#     # ensure initialized
#     # ensure registered
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_local_named_duplicate(script_runner: ScriptRunner, fake_config_home: Path,
#         fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init', str(fake_environment_directory), '--name', 'local')
#     # ensure initialized
#     # ensure registered
#     assert not ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_default(script_runner: ScriptRunner, fake_config_home: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_default_directory_exists(script_runner: ScriptRunner, fake_config_home: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init')
#     assert not ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_default_name_exists(script_runner: ScriptRunner, fake_config_home: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init')
#     assert not ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_managed(script_runner: ScriptRunner, fake_config_home: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init', '--name', 'my_env')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_init_managed_invalid_name(script_runner: ScriptRunner, fake_config_home: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'init' '--name', 'foo bar')
#     assert not ret.success
#
# # TODO: init cmdline flags for create venv, run setup?, template, list templates
# # mlonmcu validate? (env directories and environment file + user config consistency)
#
# NOP? -> mock task/dependency stuff with a few test tasks
# @pytest.mark.script_launch_mode('subprocess')
# def test_setup_default(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'setup')
#     assert ret.success
#
# NOP? -> mock task/dependency stuff with a few test tasks
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_default(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_by_name(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', '--home', 'my_env')
#     assert ret.success
#
# # TODO: --home -H -> --env -E?
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_by_path_directory(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', '--home', str(fake_environment_directory))
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_by_path_file(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', '--home', str(fake_environment_directory))
#     assert ret.success
#
# # TODO: PATH > NAME > CWD > ENVVAR > DEFAULT
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_by_env_var(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_by_cwd(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_load(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'load')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_load(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'build')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_load(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'compile')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_load(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'run')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_test(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'test')
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_test_batch(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'test', 'modelA', 'modelB', '--backend' 'backendA', '--backend',
#         'backendB', '--target', 'targetA', '--target', 'targetB')
#     # TODO: create helper function for creating test batch or single
#     assert ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_continue(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'load', ...)
#     assert ret.success
#     ret = script_runner.run('mlonmcu', 'flow', 'build', '--continue')
#     assert ret.success
#     # assert same session
#     ret = script_runner.run('mlonmcu', 'flow', 'compile', '--continue')
#     assert ret.success
#     # assert same session
#     ret = script_runner.run('mlonmcu', 'flow', 'run', '--continue')
#     assert ret.success
#     # assert same session
#     ret = script_runner.run('mlonmcu', 'flow', 'test', '--continue')
#     assert ret.success
#     # assert same session
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_continue_incompatible(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'build', ...)
#     assert ret.success
#     # assert same session
#     ret = script_runner.run('mlonmcu', 'flow', 'compile', '--continue', '--feature', 'featureA')
#     assert not ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_flow_continue_missing_parent(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'flow', 'compile', '--continue')
#     assert not ret.success
#
# @pytest.mark.script_launch_mode('subprocess')
# def test_export(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'export')
#     assert ret.success
# mlonmcu export -> Cleanup?
# test_export_list
# test_export_session_list
# test_export_run_list
# test_export_session
# test_export_run
# test_export_invalid_session
# test_export_invalid_run
# test_export_zip
# test_export_dir


# @pytest.mark.script_launch_mode('subprocess')
# def test_models(script_runner: ScriptRunner, fake_environment_directory: Path):
#     # _create_invalid_environments_file(fake_config_home / "environments.ini")
#     ret = script_runner.run('mlonmcu', 'models')
#     assert ret.success



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from mlonmcu.target.target import Target

from mlonmcu.logging import get_logger
from .microtvm_template_target import TemplateMicroTvmPlatformTarget

logger = get_logger()


class ZephyrMicroTvmPlatformTarget(TemplateMicroTvmPlatformTarget):
    DEFAULTS = {
        **Target.DEFAULTS,
        "extra_files_tar": None,
        "project_type": "host_driven",
        "zephyr_board": "",
        # "zephyr_base": "?",
        # "west_cmd": "?",
        "verbose": False,
        "warning_as_error": True,
        "compile_definitions": "",
        # "config_main_stack_size": None,
        "config_main_stack_size": "16384",
        "gdbserver_port": None,
        "nrfjprog_snr": None,
        "openocd_serial": None,
        "port": None,  # Workaround to overwrite esptool detection
    }
    REQUIRED = Target.REQUIRED | {"zephyr.install_dir", "zephyr.sdk_dir"}

    def __init__(self, name=None, features=None, config=None):
        super().__init__(name=name, features=features, config=config)
        self.template_path = None
        self.option_names = [
            "extra_files_tar",
            "project_type",
            "zephyr_board",
            # "verbose",
            "warning_as_error",
            "compile_definitions",
            "config_main_stack_size",
            "gdbserver_port",
            "nrfjprog_snr",
            "openocd_serial",
        ]
        # self.platform = platform
        # self.template = name2template(name)

    @property
    def zephyr_install_dir(self):
        return Path(self.config["zephyr.install_dir"])

    @property
    def port(self):
        return self.config["port"]

    @property
    def zephyr_sdk_dir(self):
        return Path(self.config["zephyr.sdk_dir"])

    def get_project_options(self):
        ret = super().get_project_options()
        ret.update({"zephyr_base": self.zephyr_install_dir / "zephyr"})
        return ret

    def update_environment(self, env):
        super().update_environment(env)
        env["ZEPHYR_BASE"] = str(self.zephyr_install_dir / "zephyr")
        env["ZEPHYR_SDK_INSTALL_DIR"] = str(self.zephyr_sdk_dir)
        if self.port:
            env["ESPTOOL_PORT"] = self.port



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import sys
import io
import tempfile
import json
from pathlib import Path
import tarfile
from typing import Tuple

from .tvmrt import TVMRTBackend
from mlonmcu.flow.backend import main
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.setup import utils


class TVMCGBackend(TVMRTBackend):
    name = "tvmcg"

    REQUIRED = TVMRTBackend.REQUIRED | {"utvmcg.exe"}

    def get_max_workspace_size_from_metadata(self, metadata):
        max_workspace = 0
        if "modules" in metadata:
            temp = metadata["modules"]["default"]["memory"]
        else:
            # backwards compatibility
            temp = metadata["memory"]
        for op in temp["functions"]["operator_functions"]:
            max_workspace = max(
                max_workspace,
                op["workspace"][0]["workspace_size_bytes"] if len(op["workspace"]) > 0 else 0,
            )
        return max_workspace

    def generate(self) -> Tuple[dict, dict]:
        super().generate()
        artifacts = self.artifacts
        artifact = None
        for artifact in artifacts:
            if artifact.fmt == ArtifactFormat.MLF:
                with tempfile.TemporaryDirectory() as temp_dir:
                    out_file = Path(temp_dir) / "staticrt.c"
                    mlf_path = Path(temp_dir) / "mlf"
                    mlf_bytes = io.BytesIO(artifact.raw)
                    tarfile.open(fileobj=mlf_bytes, mode="r").extractall(mlf_path)
                    artifact.export(mlf_path, extract=True)
                    metadata = None
                    with open(mlf_path / "metadata.json") as handle:
                        metadata = json.load(handle)
                    tvmcg_exe = self.config["utvmcg.exe"]
                    # graph_json_file = mlf_path / "executor-config" / "graph" / "graph.json"
                    graph_json_file = mlf_path / "executor-config" / "graph" / "default.graph"
                    params_bin_file = mlf_path / "parameters" / "default.params"
                    max_workspace_size = self.get_max_workspace_size_from_metadata(metadata)
                    args = []
                    args.append(graph_json_file)
                    args.append(params_bin_file)
                    args.append(out_file)
                    args.append(str(max_workspace_size))
                    out = utils.exec_getout(tvmcg_exe, *args, live=self.print_outputs)
                    codegen_src = open(out_file, "r").read()
                    artifact = Artifact("staticrt.c", content=codegen_src, fmt=ArtifactFormat.SOURCE)
                    workspace_size_artifact = Artifact(
                        "tvmcg_workspace_size.txt", content=f"{max_workspace_size}", fmt=ArtifactFormat.TEXT
                    )
                    artifacts.append(workspace_size_artifact)
                    stdout_artifact = Artifact("tvmcg_out.log", content=out, fmt=ArtifactFormat.TEXT)
                    artifacts.append(stdout_artifact)
                break
        assert artifact is not None, "Failed to find MLF artifact"
        artifacts.append(artifact)
        self.artifacts = artifacts


if __name__ == "__main__":
    sys.exit(
        main(
            TVMCGBackend,
            args=sys.argv[1:],
        )
    )  # pragma: no cover



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mlonmcu.target.target import Target

from mlonmcu.logging import get_logger
from mlonmcu.utils import filter_none

logger = get_logger()


class TemplateMicroTvmPlatformTarget(Target):
    REQUIRED = Target.REQUIRED | {"tvm.build_dir"}

    def __init__(self, name=None, features=None, config=None):
        super().__init__(name=name, features=features, config=config)
        self.template_path = None
        self.option_names = []

    def get_project_options(self):
        def _bool_helper(x):
            return str(x).lower() if isinstance(x, bool) else x

        return filter_none(
            {
                key: _bool_helper(getattr(self, key) if hasattr(self, key) else self.config.get(key, None))
                for key in self.option_names
            }
        )

    def update_environment(self, env):
        pass



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path

from mlonmcu.target.target import Target

from mlonmcu.logging import get_logger

from .microtvm_template_target import TemplateMicroTvmPlatformTarget

logger = get_logger()


class ArduinoMicroTvmPlatformTarget(TemplateMicroTvmPlatformTarget):
    DEFAULTS = {
        **Target.DEFAULTS,
        "project_type": None,
        "warning_as_error": False,
        "arduino_board": "?",
        # "arduino_cli_cmd": None,
        "verbose": False,
        "port": -1,
    }
    REQUIRED = Target.REQUIRED | {"arduino.install_dir"}

    def __init__(self, name=None, features=None, config=None):
        super().__init__(name=name, features=features, config=config)
        self.template_path = None
        # self.option_names = ["project_type", "warning_as_error", "arduino_board", "verbose", "port"]
        self.option_names = ["project_type", "warning_as_error", "arduino_board", "port"]
        # self.platform = platform
        # self.template = name2template(name)

    @property
    def arduino_install_dir(self):
        return Path(self.config["arduino.install_dir"])

    @property
    def port(self):
        return Path(self.config["port"])

    def get_project_options(self):
        ret = super().get_project_options()
        ret.update({"arduino_cli_cmd": self.arduino_install_dir / "arduino-cli"})
        return ret

    def update_environment(self, env):
        super().update_environment(env)
        if self.port:
            env["ESPTOOL_PORT"] = self.port  # TODO: required?



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU Target definitions"""

import os
import re
import tempfile
import time
from pathlib import Path
from typing import List, Tuple


from mlonmcu.config import filter_config
from mlonmcu.utils import filter_none
from mlonmcu.feature.feature import Feature
from mlonmcu.feature.type import FeatureType
from mlonmcu.feature.features import get_matching_features
from mlonmcu.artifact import Artifact, ArtifactFormat
from mlonmcu.config import str2bool


from mlonmcu.setup.utils import execute
from .metrics import Metrics


class Target:
    """Base target class

    Attributes
    ----------
    name : str
        Default name of the target
    features : list
        List of target features which should be enabled
    config : dict
        User config defined via key-value pairs
    inspect_program : str
        Program which can be used to inspect executables (i.e. readelf)
    inspect_program_args : list
        List of additional arguments to the inspect_program
    env : os._Environ
        Optinal map of environment variables
    """

    FEATURES = {"benchmark"}
    DEFAULTS = {
        "print_outputs": False,
        "repeat": None,
    }

    REQUIRED = set()
    OPTIONAL = set()

    def __init__(
        self,
        name: str,
        features: List[Feature] = None,
        config: dict = None,
    ):
        self.name = name
        self.config = config if config else {}
        self.pre_callbacks = []
        self.post_callbacks = []
        self.features = self.process_features(features)
        self.config = filter_config(self.config, self.name, self.DEFAULTS, self.OPTIONAL, self.REQUIRED)
        self.inspect_program = "readelf"
        self.inspect_program_args = ["--all"]
        self.env = os.environ
        self.artifacts = []
        self.dir = None

    # def init_directory(self, path=None, context=None):
    #     # return False
    #     assert path is not None
    #         self.dir = Path(path)
    #     self.dir.mkdir(exist_ok=True)

    @property
    def print_outputs(self):
        value = self.config["print_outputs"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def repeat(self):
        return self.config["repeat"]

    def __repr__(self):
        return f"Target({self.name})"

    def process_features(self, features):
        if features is None:
            return []
        features = get_matching_features(features, FeatureType.TARGET)
        for feature in features:
            assert feature.name in self.FEATURES, f"Incompatible feature: {feature.name}"
            feature.used = True
            feature.add_target_config(self.name, self.config)
            feature.add_target_callbacks(self.name, self.pre_callbacks, self.post_callbacks)
        return features

    def reconfigure(self):
        pass

    def exec(self, program: Path, *args, cwd=os.getcwd(), **kwargs):
        """Use target to execute a executable with given arguments"""
        raise NotImplementedError

    def inspect(self, program: Path, *args, **kwargs):
        """Use target to inspect a executable"""
        return execute(self.inspect_program, program, *self.inspect_program_args, *args, **kwargs)

    def parse_exit(self, out):
        exit_code = None
        exit_match = re.search(r"MLONMCU EXIT: (.*)", out)
        if exit_match:
            exit_code = int(exit_match.group(1))
        return exit_code

    def get_metrics(self, elf, directory, *args, handle_exit=None):
        # This should not be accurate, just a fallback which should be overwritten
        start_time = time.time()

        def _handle_exit(code, out=None):
            assert out is not None
            temp = self.parse_exit(out)
            # TODO: before or after?
            if temp is None:
                temp = code
            if handle_exit is not None:
                temp = handle_exit(temp, out=out)
            return temp

        if self.print_outputs:
            out, artifacts = self.exec(elf, *args, cwd=directory, live=True, handle_exit=_handle_exit)
        else:
            out, artifacts = self.exec(
                elf, *args, cwd=directory, live=False, print_func=lambda *args, **kwargs: None, handle_exit=_handle_exit
            )
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        # size instead of readelf?
        metrics = Metrics()
        metrics.add("Runtime [s]", diff)

        return metrics, out, artifacts

    def generate(self, elf) -> Tuple[dict, dict]:
        artifacts = []
        metrics = []
        total = 1 + (self.repeat if self.repeat else 0)
        # We only save the stdout and artifacts of the last execution
        # Callect metrics from all runs to aggregate them in a callback with high priority
        artifacts_ = []
        # if self.dir is None:
        #    self.dir = Path(
        with tempfile.TemporaryDirectory() as temp_dir:
            for n in range(total):
                args = []
                for callback in self.pre_callbacks:
                    callback(temp_dir, args, directory=temp_dir)
                if n == total - 1:
                    temp_dir_ = temp_dir
                else:
                    temp_dir_ = Path(temp_dir) / str(n)
                    temp_dir_.mkdir()
                metrics_, out, artifacts_ = self.get_metrics(elf, temp_dir, *args)
                metrics.append(metrics_)
            for callback in self.post_callbacks:
                out = callback(out, metrics, artifacts_, directory=temp_dir)
        artifacts.extend(artifacts_)
        if len(metrics) > 1:
            raise RuntimeError("Collected target metrics for multiple runs. Please aggregate them in a callback!")
        assert len(metrics) == 1
        metrics = metrics[0]
        artifacts_ = {"default": artifacts}
        if not isinstance(metrics, dict):
            metrics = {"default": metrics}
        stdout_artifact = Artifact(
            f"{self.name}_out.log", content=out, fmt=ArtifactFormat.TEXT
        )  # TODO: rename to tvmaot_out.log?
        artifacts_["default"].append(stdout_artifact)
        return artifacts_, metrics

    def generate_artifacts(self, elf):
        start_time = time.time()
        artifacts, metrics = self.generate(elf)
        # TODO: do something with out?
        end_time = time.time()
        diff = end_time - start_time
        for name, metrics_ in metrics.items():
            if name == "default":
                metrics_.add("Run Stage Time [s]", diff, True)
            content = metrics_.to_csv(include_optional=True)  # TODO: store df instead?
            artifact = Artifact("run_metrics.csv", content=content, fmt=ArtifactFormat.TEXT, flags=["metrics"])
            # Alternative: artifact = Artifact("metrics.csv", data=df/dict, fmt=ArtifactFormat.DATA)
            if name not in artifacts:
                artifacts[name] = []
            artifacts[name].append(artifact)

        self.artifacts = artifacts
        return artifacts

    def export_artifacts(self, path):
        assert len(self.artifacts) > 0, "No artifacts found, please run generate_artifacts() first"

        if not isinstance(path, Path):
            path = Path(path)

        is_dir = len(path.suffix) == 0
        if is_dir:
            assert (
                path.is_dir()
            ), "The supplied path does not exists."  # Make sure it actually exists (we do not create it by default)
            for artifact in self.artifacts:
                artifact.export(path)
        else:
            raise NotImplementedError

    def get_target_system(self):
        return self.name

    def get_arch(self):
        raise NotImplementedError

    def get_platform_config(self, platform):
        return {}

    def add_platform_config(self, platform, config):
        config.update(self.get_platform_config(platform))

    def get_platform_defs(self, platform):
        return {}

    def add_platform_defs(self, platform, defs):
        defs.update(self.get_platform_defs(platform))

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        return {}

    def add_backend_config(self, backend, config, optimized_layouts=False, optimized_schedules=False):
        new = filter_none(
            self.get_backend_config(
                backend, optimized_layouts=optimized_layouts, optimized_schedules=optimized_schedules
            )
        )

        # only allow overwriting non-none values
        # to support accepting user-vars
        new = {key: value for key, value in new.items() if config.get(key, None) is None}
        config.update(new)

    def get_hardware_details(self):
        return {
            "num-cores": 1,  # TODO: overwrite for host_x86?
            "cache-line-bytes": 64,  # TODO: disable?
            "vector-unit-bytes": 64,  # TODO: disable?
            # The following are GPU specific
            "max-shared-memory-per-block": 0,
            "max-local-memory-per-block": 0,
            "max-threads-per-block": 0,
            "max-vthread-extent": 0,
            "warp-size": 0,
        }



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for installing a mlonmcu environment."""
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import setup

from mlonmcu.cli.common import (
    add_common_options,
    add_context_options,
)

# from .helper.parse import extract_config_and_init_features
from .helper.parse import extract_config_and_feature_names


def add_setup_options(parser):
    setup_parser = parser.add_argument_group("setup options")
    setup_parser.add_argument(
        "-f",
        "--feature",
        type=str,
        metavar="FEATURE",
        # nargs=1,
        action="append",
        choices=[],  # FIXME: only setup features?
        help="Enabled features for setup (choices: %(choices)s)",
    )
    setup_parser.add_argument(
        "-c",
        "--config",
        metavar="KEY=VALUE",
        nargs="+",
        action="append",
        help=(
            "Set a number of key-value pairs "
            "(do not put spaces before or after the = sign). "
            "If a value contains spaces, you should define "
            "it with double quotes: "
            'foo="this is a sentence". Note that '
            "values are always treated as strings."
        ),
    )
    setup_parser.add_argument(
        "-p",
        "--progress",
        action="store_true",
        help="Display progress bar (default: %(default)s)",
    )
    setup_parser.add_argument(
        "-r",
        "--rebuild",
        action="store_true",
        help="Trigger a rebuild/refresh of already installed dependencies (default: %(default)s)",
    )
    setup_parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="Only print a list of the tasks to be processed and quit (default: %(default)s)",
    )
    setup_parser.add_argument(
        "-g",
        "--generate-requirements",
        action="store_true",
        help="Generate requirements_addition.txt listing the dependent python packages.",
    )
    setup_parser.add_argument(
        "--task",
        type=str,
        nargs=1,
        default=None,
        help="Invoke a single task manually by name (default: %(default)s)",
    )
    setup_parser.add_argument(
        "--visualize",
        type=str,
        nargs=1,
        help="Export a .dot file for the dependency graph.",
    )


def get_parser(subparsers):
    """ "Define and return a subparser for the setup subcommand."""
    parser = subparsers.add_parser("setup", description="Setup ML on MCU dependencies.")
    parser.set_defaults(func=handle)
    add_common_options(parser)
    add_context_options(parser)
    add_setup_options(parser)
    return parser


def handle(args):
    with MlonMcuContext(path=args.home, deps_lock="write") as context:
        # config, features = extract_config_and_init_features(args)
        config, _, _, _ = extract_config_and_feature_names(args)
        # installer = setup.Setup(features=features, config=config, context=context)
        installer = setup.Setup(config=config, context=context)
        if args.visualize:
            installer.visualize(args.visualize[0], ordered=False)
        elif args.list:
            order = installer.get_dependency_order()
            print("The following tasks have been selected:")
            print("\n".join(["- " + key for key in order]))
        elif args.task:
            installer.invoke_single_task(
                args.task[0],
                progress=args.progress,
                rebuild=args.rebuild,
            )
        elif args.generate_requirements:
            installer.generate_requirements()
            print(
                "Next you should run `python -m pip install -r $MLONMCU_HOME/requirements_addition.txt` inside your"
                " virtual environment."
            )
        else:
            installer.install_dependencies(
                progress=args.progress,
                rebuild=args.rebuild,
            )



# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .riscv import (
    EtissPulpinoTarget,
    SpikeTarget,
    OVPSimTarget,
    COREVOVPSimTarget,
    RiscvQemuTarget,
    GvsocPulpTarget,
    EtissTarget,
    AraTarget,
    AraRtlTarget,
    CV32E40PTarget,
    VicunaTarget,
)
from .arm import Corstone300Target
from .host_x86 import HostX86Target

TARGET_REGISTRY = {}


def register_target(target_name, t, override=False):
    global TARGET_REGISTRY

    if target_name in TARGET_REGISTRY and not override:
        raise RuntimeError(f"Target {target_name} is already registered")
    TARGET_REGISTRY[target_name] = t


def get_targets():
    return TARGET_REGISTRY


register_target("etiss_pulpino", EtissPulpinoTarget)
register_target("etiss", EtissTarget)
register_target("host_x86", HostX86Target)
register_target("corstone300", Corstone300Target)
register_target("spike", SpikeTarget)
register_target("ovpsim", OVPSimTarget)
register_target("corev_ovpsim", COREVOVPSimTarget)
register_target("riscv_qemu", RiscvQemuTarget)
register_target("gvsoc_pulp", GvsocPulpTarget)
register_target("ara", AraTarget)
register_target("ara_rtl", AraRtlTarget)
register_target("cv32e40p", CV32E40PTarget)
register_target("vicuna", VicunaTarget)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .tf import _validate_tensorflow
from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_tflite_micro_compiler(context: MlonMcuContext, params=None):
    if not _validate_tensorflow(context, params=params):
        return False
    if not context.environment.has_backend("tflmc"):
        return False
    return True


@Tasks.provides(["tflmc.src_dir"])
@Tasks.validate(_validate_tflite_micro_compiler)
@Tasks.register(category=TaskType.BACKEND)
def clone_tflite_micro_compiler(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Clone the preinterpreter repository."""
    tflmcName = utils.makeDirName("tflmc")
    tflmcSrcDir = context.environment.paths["deps"].path / "src" / tflmcName
    if rebuild or not utils.is_populated(tflmcSrcDir):
        tflmcRepo = context.environment.repos["tflite_micro_compiler"]
        utils.clone_wrapper(tflmcRepo, tflmcSrcDir, refresh=rebuild)
    context.cache["tflmc.src_dir"] = tflmcSrcDir


def _validate_build_tflite_micro_compiler(context: MlonMcuContext, params=None):
    if params:
        muriscvnn = params.get("muriscvnn", False)
        cmsisnn = params.get("cmsisnn", False)
        if muriscvnn and cmsisnn:
            # Not allowed
            return False
        elif muriscvnn:
            if not context.environment.supports_feature("muriscvnn"):
                return False
        elif cmsisnn:
            if not context.environment.supports_feature("cmsisnn"):
                return False
    return _validate_tflite_micro_compiler(context, params=params)


@Tasks.needs(["tflmc.src_dir", "tf.src_dir"])
@Tasks.optional(["muriscvnn.lib", "muriscvnn.inc_dir", "cmsisnn.dir"])
@Tasks.provides(["tflmc.build_dir", "tflmc.exe"])
@Tasks.param("muriscvnn", [False])
@Tasks.param("cmsisnn", [False, True])
@Tasks.param("dbg", [False, True])
@Tasks.param("arch", ["x86"])  # TODO: compile for arm/riscv in the future
@Tasks.validate(_validate_build_tflite_micro_compiler)
@Tasks.register(category=TaskType.BACKEND)
def build_tflite_micro_compiler(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Build the TFLM preinterpreter."""
    muriscvnn = params.get("muriscvnn", False)
    cmsisnn = params.get("cmsisnn", False)
    dbg = params.get("dbg", False)
    arch = params.get("arch", "x86")
    flags = utils.makeFlags((True, arch), (muriscvnn, "muriscvnn"), (cmsisnn, "cmsisnn"), (dbg, "dbg"))
    # flags_ = utils.makeFlags((dbg, "dbg"))
    tflmcName = utils.makeDirName("tflmc", flags=flags)
    tflmcBuildDir = context.environment.paths["deps"].path / "build" / tflmcName
    tflmcInstallDir = context.environment.paths["deps"].path / "install" / tflmcName
    tflmcExe = tflmcInstallDir / "compiler"
    tfSrcDir = context.cache["tf.src_dir"]
    tflmcSrcDir = context.cache["tflmc.src_dir"]
    if rebuild or not utils.is_populated(tflmcBuildDir) or not tflmcExe.is_file():
        cmakeArgs = [
            "-DTF_SRC=" + str(tfSrcDir),
            "-DGET_TF_SRC=ON",
        ]
        if muriscvnn:
            flags__ = utils.makeFlags((True, arch), (True, "gcc"), (dbg, "dbg"))
            muriscvnnLib = context.cache["muriscvnn.lib", flags__]
            muriscvnnInc = context.cache["muriscvnn.inc_dir"]
            cmakeArgs.append("-DTFLM_OPTIMIZED_KERNEL=cmsis_nn")
            cmakeArgs.append(f"-DTFLM_OPTIMIZED_KERNEL_LIB={muriscvnnLib}")
            cmakeArgs.append(f"-DTFLM_OPTIMIZED_KERNEL_INCLUDE_DIR={muriscvnnInc}")
        elif cmsisnn:
            flags__ = utils.makeFlags((True, arch), (dbg, "dbg"))
            cmsisnnLib = context.cache["cmsisnn.lib", flags__]
            cmsisDir = Path(context.cache["cmsisnn.dir"])
            cmsisIncs = r"\;".join(
                [
                    str(cmsisDir),
                    str(cmsisDir / "CMSIS" / "Core" / "Include"),
                    str(cmsisDir / "CMSIS" / "NN" / "Include"),
                    str(cmsisDir / "CMSIS" / "DSP" / "Include"),
                ]
            )
            cmakeArgs.append("-DTFLM_OPTIMIZED_KERNEL=cmsis_nn")
            cmakeArgs.append(f"-DTFLM_OPTIMIZED_KERNEL_LIB={cmsisnnLib}")
            cmakeArgs.append(f"-DTFLM_OPTIMIZED_KERNEL_INCLUDE_DIR={cmsisIncs}")
        utils.mkdirs(tflmcBuildDir)
        # utils.cmake("-DTF_SRC=" + str(tfSrcDir), str(tflmcSrcDir), debug=params["dbg"], cwd=tflmcBuildDir)
        utils.cmake(
            *cmakeArgs,
            str(tflmcSrcDir),
            debug=dbg,
            cwd=tflmcBuildDir,
            live=verbose,
        )
        utils.make(cwd=tflmcBuildDir, threads=threads, live=verbose)
        utils.mkdirs(tflmcInstallDir)
        utils.move(tflmcBuildDir / "compiler", tflmcExe)
    context.cache["tflmc.build_dir", flags] = tflmcBuildDir
    context.cache["tflmc.exe", flags] = tflmcExe
    context.export_paths.add(tflmcExe.parent)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
from collections import ChainMap

NUM_GEN_ARGS = 9


def parse_var(s):
    """
    Parse a key, value pair, separated by '='
    That's the reverse of ShellArgs.

    On the command line (argparse) a declaration will typically look like:
        foo=hello
    or
        foo="hello world"
    """
    items = s.split("=")
    key = items[0].strip()  # we remove blanks around keys, as is logical
    if len(key) == 0 or len(items) <= 1 or "=" not in s:
        raise RuntimeError(f"The argument {s} is not a key-value pair")
    if len(items) > 1:
        # rejoin the rest:
        value = "=".join(items[1:])
    return (key, value)


def parse_vars(items):  # TODO: this needs to be used in other subcommands as well?
    """
    Parse a series of key-value pairs and return a dictionary
    """
    d = {}

    if items:
        for item in items:
            if len(item) > 0:
                key, value = parse_var(item)
                d[key] = value
    return d


def extract_feature_names(args):
    if args.feature:
        features = args.feature
    else:
        features = []

    def helper(args, name):
        gen = []
        if hasattr(args, name):
            feature_gen = getattr(args, name)
            if not feature_gen:
                gen = [[]]
                return gen
            for x in feature_gen:
                if "_" in x:
                    assert len(set(x)) == 1
                    gen.append([])
                else:
                    gen.append(x)
        else:
            gen = [[]]
        return gen

    gens = []
    for i in range(NUM_GEN_ARGS):
        suffix = str(i + 1) if i > 0 else ""
        gen = helper(args, "feature_gen" + suffix)
        gens.append(gen)

    gen = list(map(lambda x: sum(x, []), (itertools.product(*gens))))

    return features, gen


def extract_config(args):
    if args.config:
        configs = sum(args.config, [])
        configs = parse_vars(configs)
    else:
        configs = {}

    def helper(args, name):
        gen = []
        if hasattr(args, name):
            config_gen = getattr(args, name)
            if not config_gen:
                gen = [{}]
                return gen
            for x in config_gen:
                if "_" in x:
                    assert len(set(x)) == 1
                    gen.append({})
                else:
                    c = parse_vars(x)
                    gen.append(c)
        else:
            gen = [{}]
        return gen

    gens = []
    for i in range(NUM_GEN_ARGS):
        suffix = str(i + 1) if i > 0 else ""
        gen = helper(args, "config_gen" + suffix)
        gens.append(gen)

    gen = list(map(lambda x: dict(ChainMap(*x)), (itertools.product(*gens))))

    return configs, gen


def extract_config_and_feature_names(args, context=None):
    # TODO: get features from context?
    feature_names, feature_gen = extract_feature_names(args)
    config, config_gen = extract_config(args)
    return config, feature_names, config_gen, feature_gen


def extract_frontend_names(args, context=None):
    frontend_names = args.frontend
    names = []
    if isinstance(frontend_names, list) and len(frontend_names) > 0:
        names = frontend_names
    elif isinstance(frontend_names, str):
        names = [frontend_names]
    else:
        # No need to specify a default, because we just use the provided order in the environment.yml
        assert frontend_names is None, "TODO"
        assert context is not None, "Need context to resolve default frontends"
        all_frontend_names = context.environment.lookup_frontend_configs(names_only=True)
        names.extend(all_frontend_names)
    return names


def extract_postprocess_names(args, context=None):
    return list(dict.fromkeys(args.postprocess)) if args.postprocess is not None else []


def extract_backend_names(args, context=None):
    if isinstance(args.backend, list) and len(args.backend) > 0:
        backends = args.backend
    elif isinstance(args.backend, str):
        backends = [args.backend]
    else:
        assert args.backend is None, "TODO"
        assert context is not None
        frameworks = context.environment.get_default_frameworks()
        backends = []
        for framework in frameworks:
            framework_backends = context.environment.get_default_backends(framework)
            backends.extend(framework_backends)
    return backends


def extract_target_names(args, context=None):
    if isinstance(args.target, list) and len(args.target) > 0:
        targets = args.target
    elif isinstance(args.target, str):
        targets = [args.target]
    else:
        assert args.target is None, "TODO"
        # assert context is not None
        if context is None:
            return [None]
        targets = context.environment.get_default_targets()
    return targets


def extract_platform_names(args, context=None):
    if args.platform:
        platforms = [x[0] for x in args.platform]
    else:
        assert args.platform is None
        if context is None:
            return [None]
        platforms = context.environment.lookup_platform_configs(names_only=True)
    return platforms



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of a MLonMCU Run which represents a set of benchmarks in a session."""
import os
import shutil
import filelock
import tempfile
import multiprocessing
from datetime import datetime
from enum import Enum
from pathlib import Path
import concurrent.futures

from tqdm import tqdm

from mlonmcu.session.run import Run
from mlonmcu.logging import get_logger
from mlonmcu.report import Report
from mlonmcu.config import filter_config

from .postprocess.postprocess import SessionPostprocess
from .run import RunStage

logger = get_logger()  # TODO: rename to get_mlonmcu_logger


class SessionStatus(Enum):  # TODO: remove?
    """Status type for a session."""

    CREATED = 0
    OPEN = 1
    CLOSED = 2
    ERROR = 3


class Session:
    """A session which wraps around multiple runs in a context."""

    DEFAULTS = {
        "report_fmt": "csv",
    }

    def __init__(self, label="", idx=None, archived=False, dir=None, config=None):
        self.timestamp = datetime.now().strftime("%Y%m%dT%H%M%S")
        self.label = (
            label if len(label) > 0 else ("unnamed" + "_" + self.timestamp)
        )  # TODO: decide if named sessions should also get a timestamp?
        self.idx = idx
        self.config = config if config else {}
        self.config = filter_config(self.config, "session", self.DEFAULTS, set(), set())
        self.status = SessionStatus.CREATED
        self.opened_at = None
        self.closed_at = None
        self.runs = []
        self.report = None
        self.next_run_idx = 0
        self.archived = archived
        self.dir = dir
        self.tempdir = None
        self.session_lock = None

    @property
    def runs_dir(self):
        return None if self.dir is None else (self.dir / "runs")

    def __enter__(self):
        if self.archived:
            logger.warning("Opening an already archived session is not recommended")
        else:
            self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.active:
            self.close()

    @property
    def prefix(self):
        """get prefix property."""
        return f"[session-{self.idx}] " if self.idx else ""

    @property
    def report_fmt(self):
        """get report_fmt property."""
        return str(self.config["report_fmt"])

    def create_run(self, *args, **kwargs):
        """Factory method to create a run and add it to this session."""
        idx = len(self.runs)
        logger.debug("Creating a new run with id %s", idx)
        run = Run(*args, idx=idx, session=self, **kwargs)
        self.runs.append(run)
        return run

    #  def update_run(self): # TODO TODO
    #      pass

    def get_reports(self):
        """Returns a full report which includes all runs in this session."""
        if self.report:
            return self.report

        reports = [run.get_report() for run in self.runs]
        merged = Report()
        merged.add(reports)
        return merged

    def enumerate_runs(self):
        """Update run indices."""
        # Find start index
        max_idx = -1
        for run in self.runs:
            if run.archived:
                max_idx = max(max_idx, run.idx)
        run_idx = max_idx + 1
        last_run_idx = None
        for run in self.runs:
            if not run.archived:
                run.idx = run_idx
                run.init_directory()
                run_idx += 1
                last_run_idx = run.idx
        self.next_run_idx = run_idx
        if last_run_idx is not None:
            self.update_latest_run_symlink(last_run_idx)

    def update_latest_run_symlink(self, latest_run_idx):
        run_link = self.runs_dir / "latest"  # TODO: Create relative symlink using os.path.relpath for portability
        if os.path.islink(run_link):
            os.unlink(run_link)
        os.symlink(self.runs_dir / str(latest_run_idx), run_link)

    def request_run_idx(self):
        """Return next free run index."""
        ret = self.next_run_idx
        self.next_run_idx += 1
        # TODO: find a better approach for this
        return ret

    def process_runs(
        self,
        until=RunStage.DONE,
        per_stage=False,
        print_report=False,
        num_workers=1,
        progress=False,
        export=False,
        context=None,
    ):
        """Process a runs in this session until a given stage."""

        # TODO: Add configurable callbacks for stage/run complete
        assert self.active, "Session needs to be opened first"

        self.enumerate_runs()
        self.report = None
        assert num_workers > 0, "num_workers can not be < 1"
        workers = []
        # results = []
        workers = []
        pbar = None  # Outer progress bar
        pbar2 = None  # Inner progress bar
        num_runs = len(self.runs)
        num_failures = 0
        stage_failures = {}
        worker_run_idx = []

        def _init_progress(total, msg="Processing..."):
            """Helper function to initialize a progress bar for the session."""
            return tqdm(
                total=total,
                desc=msg,
                ncols=100,
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}s]",
                leave=None,
            )

        def _update_progress(pbar, count=1):
            """Helper function to update the progress bar for the session."""
            pbar.update(count)

        def _close_progress(pbar):
            """Helper function to close the session progressbar, if available."""
            if pbar:
                pbar.close()

        def _process(pbar, run, until, skip):
            """Helper function to invoke the run."""
            run.process(until=until, skip=skip, export=export)
            if progress:
                _update_progress(pbar)

        def _join_workers(workers):
            """Helper function to collect all worker threads."""
            nonlocal num_failures
            results = []
            for i, w in enumerate(workers):
                try:
                    results.append(w.result())
                except Exception as e:
                    logger.exception(e)
                    logger.error("An exception was thrown by a worker during simulation")
                run_index = worker_run_idx[i]
                run = self.runs[run_index]
                if run.failing:
                    num_failures += 1
                    failed_stage = RunStage(run.next_stage).name
                    if failed_stage in stage_failures:
                        stage_failures[failed_stage].append(run_index)
                    else:
                        stage_failures[failed_stage] = [run_index]
            if progress:
                _close_progress(pbar)
            return results

        def _used_stages(runs, until):
            """Determines the stages which are used by at least one run."""
            used = []
            for stage_index in list(range(RunStage.LOAD, until + 1)) + [RunStage.POSTPROCESS]:
                stage = RunStage(stage_index)
                if any(run.has_stage(stage) for run in runs):
                    used.append(stage)
            return used

        used_stages = _used_stages(self.runs, until)
        skipped_stages = [stage for stage in RunStage if stage not in used_stages]

        with concurrent.futures.ThreadPoolExecutor(num_workers) as executor:
            if per_stage:
                if progress:
                    pbar2 = _init_progress(len(used_stages), msg="Processing stages")
                for stage in used_stages:
                    run_stage = RunStage(stage).name
                    if progress:
                        pbar = _init_progress(len(self.runs), msg=f"Processing stage {run_stage}")
                    else:
                        logger.info("%s Processing stage %s", self.prefix, run_stage)
                    for i, run in enumerate(self.runs):
                        if i == 0:
                            total_threads = min(len(self.runs), num_workers)
                            cpu_count = multiprocessing.cpu_count()
                            if (stage == RunStage.COMPILE) and run.compile_platform:
                                total_threads *= run.compile_platform.num_threads
                            if total_threads > 2 * cpu_count:
                                if pbar2:
                                    print()
                                logger.warning(
                                    "The chosen configuration leads to a maximum of %d threads being"
                                    + " processed which heavily exceeds the available CPU resources (%d)."
                                    + " It is recommended to lower the value of 'mlif.num_threads'!",
                                    total_threads,
                                    cpu_count,
                                )
                        if run.failing:
                            logger.warning("Skiping stage '%s' for failed run", run_stage)
                        else:
                            worker_run_idx.append(i)
                            workers.append(executor.submit(_process, pbar, run, until=stage, skip=skipped_stages))
                    _join_workers(workers)
                    workers = []
                    worker_run_idx = []
                    if progress:
                        _update_progress(pbar2)
                if progress:
                    _close_progress(pbar2)
            else:
                if progress:
                    pbar = _init_progress(len(self.runs), msg="Processing all runs")
                else:
                    logger.info(self.prefix + "Processing all stages")
                for i, run in enumerate(self.runs):
                    if i == 0:
                        total_threads = min(len(self.runs), num_workers)
                        cpu_count = multiprocessing.cpu_count()
                        if (
                            (until >= RunStage.COMPILE)
                            and run.compile_platform is not None
                            and run.compile_platform.name == "mlif"
                        ):
                            total_threads *= (
                                run.compile_platform.num_threads
                            )  # TODO: This should also be used for non-mlif platforms
                        if total_threads > 2 * cpu_count:
                            if pbar2:
                                print()
                            logger.warning(
                                "The chosen configuration leads to a maximum of %d being processed which"
                                + " heavily exceeds the available CPU resources (%d)."
                                + " It is recommended to lower the value of 'mlif.num_threads'!",
                                total_threads,
                                cpu_count,
                            )
                    worker_run_idx.append(i)
                    workers.append(executor.submit(_process, pbar, run, until=until, skip=skipped_stages))
                _join_workers(workers)
        if num_failures == 0:
            logger.info("All runs completed successfuly!")
        elif num_failures == num_runs:
            logger.error("All runs have failed to complete!")
        else:
            num_success = num_runs - num_failures
            logger.warning("%d out or %d runs completed successfully!", num_success, num_runs)
            summary = "\n".join(
                [
                    f"\t{stage}: \t{len(failed)} failed run(s): " + " ".join([str(idx) for idx in failed])
                    for stage, failed in stage_failures.items()
                    if len(failed) > 0
                ]
            )
            logger.info("Summary:\n%s", summary)

        report = self.get_reports()
        logger.info("Postprocessing session report")
        # Warning: currently we only support one instance of the same type of postprocess,
        # also it will be applied to all rows!
        session_postprocesses = []
        for run in self.runs:
            for postprocess in run.postprocesses:
                if isinstance(postprocess, SessionPostprocess):
                    if postprocess.name not in [p.name for p in session_postprocesses]:
                        session_postprocesses.append(postprocess)
        for postprocess in session_postprocesses:
            artifacts = postprocess.post_session(report)
            if artifacts is not None:
                for artifact in artifacts:
                    # Postprocess has an artifact: write to disk!
                    logger.debug("Writting postprocess artifact to disk: %s", artifact.name)
                    artifact.export(self.dir)
        report_file = Path(self.dir) / f"report.{self.report_fmt}"
        report.export(report_file)
        results_dir = context.environment.paths["results"].path
        results_file = results_dir / f"{self.label}.{self.report_fmt}"
        report.export(results_file)
        logger.info(self.prefix + "Done processing runs")
        self.report = report
        if print_report:
            logger.info("Report:\n%s", str(report.df))

        return num_failures == 0

    def discard(self):
        """Discard a run and remove its directory."""
        self.close()
        if self.dir.is_dir():
            logger.debug("Cleaning up discarded session")
            shutil.rmtree(self.dir)

    def __repr__(self):
        return f"Session(idx={self.idx},status={self.status},runs={self.runs})"

    @property
    def active(self):
        """Get active property."""
        return self.status == SessionStatus.OPEN

    @property
    def failing(self):
        """Get failng property."""

        # via report
        if self.report:
            df = self.report.df
            if "Failing" in df.columns:
                if df["Failing"].any():
                    return True
        # via runs
        if len(self.runs) > 0:
            for run in self.runs:
                if run.failing:
                    return True

        return False

    def open(self):
        """Open this run."""
        self.status = SessionStatus.OPEN
        self.opened_at = datetime.now()
        if dir is None:
            assert not self.archived
            self.tempdir = tempfile.TemporaryDirectory()
            self.dir = Path(self.tempdir.name)
        else:
            if not self.dir.is_dir():
                self.dir.mkdir(parents=True)
        self.session_lock = filelock.FileLock(os.path.join(self.dir, ".lock"))
        try:
            self.session_lock.acquire(timeout=10)
        except filelock.Timeout as err:
            raise RuntimeError("Lock on session could not be aquired.") from err
        if not os.path.exists(self.runs_dir):
            os.mkdir(self.runs_dir)

    def close(self, err=None):
        """Close this run."""
        if err:
            self.status = SessionStatus.ERROR
        else:
            self.status = SessionStatus.CLOSED
        self.closed_at = datetime.now()
        self.session_lock.release()
        os.remove(self.session_lock.lock_file)
        if self.tempdir:
            self.tempdir.cleanup()



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

from .config import (
    DefaultsConfig,
    PathConfig,
    RepoConfig,
    FrameworkConfig,
    FrameworkFeatureConfig,
    BackendConfig,
    BackendFeatureConfig,
    TargetConfig,
    TargetFeatureConfig,
    PlatformConfig,
    PlatformFeatureConfig,
    FrontendConfig,
    FrontendFeatureConfig,
)
from .loader import load_environment_from_file
from .writer import write_environment_to_file

from mlonmcu.feature.type import FeatureType


def _feature_helper(obj, name):
    if not obj.enabled:
        return []
    features = obj.features
    if name:
        return [feature for feature in features if feature.name == name]
    else:
        return features


def _extract_names(objs):
    return [obj.name for obj in objs]


def _filter_enabled(objs):
    return [obj for obj in objs if obj.enabled]


class Environment:
    def __init__(self):
        self._home = None
        self.alias = None
        self.defaults = DefaultsConfig()
        self.paths = {}
        self.repos = {}
        self.frameworks = []
        self.frontends = []
        self.platforms = []
        self.toolchains = []
        self.targets = []
        self.vars = {}
        self.flags = {}

    def __str__(self):
        return self.__class__.__name__ + "(" + str(vars(self)) + ")"

    @property
    def home(self):
        """Home directory of mlonmcu environment."""
        return self._home

    @classmethod
    def from_file(cls, filename):
        return load_environment_from_file(filename, base=cls)

    def to_file(self, filename):
        write_environment_to_file(self, filename)

    def lookup_path(self, name):
        assert name in self.paths, f"Unable to find '{name}' path in environment config"
        return self.paths[name]

    def lookup_var(self, name, default=None):
        return self.vars.get(name, default)

    def lookup_frontend_feature_configs(self, name=None, frontend=None):
        configs = []
        if frontend:
            names = [frontend.name for frontend in self.frontends]
            index = names.index(frontend)
            assert index is not None, f"Frontend {frontend} not found in environment config"
            configs.extend(_feature_helper(self.frontends[index], name))
        else:
            for frontend in self.frontends:
                configs.extend(_feature_helper(frontend, name))
        return configs

    def lookup_framework_feature_configs(self, name=None, framework=None):
        configs = []
        if framework:
            names = [framework.name for framework in self.frameworks]
            index = names.index(framework)
            assert index is not None, f"Framework {framework} not found in environment config"
            configs.extend(_feature_helper(self.frameworks[index], name))
        else:
            for framework in self.frameworks:
                configs.extend(_feature_helper(framework, name))
        return configs

    def lookup_backend_feature_configs(self, name=None, framework=None, backend=None):
        def helper(framework, backend, name):
            backend_features = self.framework[framework].backends[backend].features
            if name:
                return [backend_features[name]]
            else:
                return backend_features.values()

        configs = []
        if framework:
            names = [framework.name for framework in self.frameworks]
            index = names.index(framework)
            assert index is not None, f"Framework {framework} not found in environment config"
            if backend:
                names_ = [backend.name for backend in self.frameworks[index].backends]
                index_ = names_.index(backend)
                assert index_ is not None, f"Backend {backend} not found in environment config"
                configs.extend(_feature_helper(self.frameworks[index].backends[index], name))
            else:
                for backend in self.frameworks[index].backends:
                    configs.extend(_feature_helper(backend, name))
        else:
            for framework in self.frameworks:
                if backend:
                    names_ = [backend.name for backend in framework.backends]
                    index_ = names_.index(backend)
                    assert index_ is not None, f"Backend {backend} not found in environment config"
                    configs.extend(_feature_helper(self.frameworks[index].backends[index], name))
                else:
                    for backend in framework.backends:
                        configs.extend(_feature_helper(backend, name))
                    backend = None
        return configs

    def lookup_platform_feature_configs(self, name=None, platform=None):
        configs = []
        if platform:
            names = [platform.name for platform in self.platforms]
            index = names.index(platform)
            assert (
                index is not None
            ), f"Platform {platform} not found in environment config"  # TODO: do not fail, just return empty list
            configs.extend(_feature_helper(self.platforms[index], name))
        else:
            for platform in self.platforms:
                configs.extend(_feature_helper(platform, name))
        return configs

    def lookup_target_feature_configs(self, name=None, target=None):
        configs = []
        if target:
            names = [target.name for target in self.targets]
            index = names.index(target)
            assert (
                index is not None
            ), f"Target {target} not found in environment config"  # TODO: do not fail, just return empty list
            configs.extend(_feature_helper(self.targets[index], name))
        else:
            for target in self.targets:
                configs.extend(_feature_helper(target, name))
        return configs

    def lookup_feature_configs(
        self,
        name=None,
        kind=None,
        frontend=None,
        framework=None,
        backend=None,
        platform=None,
        target=None,
    ):
        configs = []
        if kind == FeatureType.FRONTEND or kind is None:
            configs.extend(self.lookup_frontend_feature_configs(name=name, frontend=frontend))
        if kind == FeatureType.FRAMEWORK or kind is None:
            configs.extend(self.lookup_framework_feature_configs(name=name, framework=framework))
        if kind == FeatureType.BACKEND or kind is None:
            configs.extend(self.lookup_backend_feature_configs(name=name, framework=framework, backend=backend))
        if kind == FeatureType.PLATFORM or kind is None:
            configs.extend(self.lookup_platform_feature_configs(name=name, platform=platform))
        if kind == FeatureType.TARGET or kind is None:
            configs.extend(self.lookup_target_feature_configs(name=name, target=target))

        return configs

    def supports_feature(self, name):
        configs = self.lookup_feature_configs(name=name)
        supported = [feature.supported for feature in configs]
        return any(supported)

    def has_feature(self, name):
        """An alias for supports_feature."""
        return self.supports_feature(name)

    def lookup_backend_configs(self, backend=None, framework=None, names_only=False):
        enabled_frameworks = _filter_enabled(self.frameworks)

        configs = []
        for framework_config in enabled_frameworks:
            if framework is not None and framework_config.name != framework:
                continue
            enabled_backends = _filter_enabled(framework_config.backends)
            if backend is None:
                configs.extend(enabled_backends)
            else:
                for backend_config in enabled_backends:
                    if backend_config.name == backend:
                        return [backend_config.name if names_only else backend_config]
        return _extract_names(configs) if names_only else configs

    def lookup_framework_configs(self, framework=None, names_only=False):
        enabled_frameworks = _filter_enabled(self.frameworks)

        if framework is None:
            return _extract_names(enabled_frameworks) if names_only else enabled_frameworks

        for framework_config in enabled_frameworks:
            if framework_config.name == framework:
                return [framework_config.name if names_only else framework_config]

        return []

    def lookup_frontend_configs(self, frontend=None, names_only=False):
        enabled_frontends = _filter_enabled(self.frontends)

        if frontend is None:
            return _extract_names(enabled_frontends) if names_only else enabled_frontends

        for frontend_config in enabled_frontends:
            if frontend_config.name == frontend:
                return [frontend_config.name if names_only else frontend_config]
        return []

    def lookup_platform_configs(self, platform=None, names_only=False):
        enabled_platforms = _filter_enabled(self.platforms)

        if platform is None:
            return _extract_names(enabled_platforms) if names_only else enabled_platforms
        for platform_config in enabled_platforms:
            if platform_config.name == platform:
                return [platform_config.name if names_only else platform_config]

        return []

    def lookup_target_configs(self, target=None, names_only=False):
        enabled_targets = _filter_enabled(self.targets)

        if target is None:
            return _extract_names(enabled_targets) if names_only else enabled_targets

        for target_config in enabled_targets:
            if target_config.name == target:
                return [target_config.name if names_only else target_config]
        return []

    def has_frontend(self, name):
        configs = self.lookup_frontend_configs(frontend=name)
        return len(configs) > 0

    def has_backend(self, name):
        configs = self.lookup_backend_configs(backend=name)
        return len(configs) > 0

    def has_framework(self, name):
        configs = self.lookup_framework_configs(framework=name)
        return len(configs) > 0

    def has_platform(self, name):
        configs = self.lookup_platform_configs(platform=name)
        return len(configs) > 0

    def has_toolchain(self, name):
        return self.toolchains.get(name, False)

    def has_target(self, name):
        configs = self.lookup_target_configs(target=name)
        return len(configs) > 0

    # TODO: actually we do not need to explicitly enable those? environment.yml list the default enabled ones instead
    # of the supported ones in the environment
    # def has_postprocess(self, name):
    #     configs = self.lookup_postprocess_configs(postprocess=name)
    #     return len(configs) > 0

    def get_default_backends(self, framework):
        if framework is None or framework not in self.defaults.default_backends:
            return []
        default = self.defaults.default_backends[framework]
        # framework_names = [framework_config.name for framework_config in self.frameworks]
        # framework_config = self.frameworks[framework_names.index(framework)]
        if default is None:
            return []
        if isinstance(default, str):
            if default == "*":  # Wildcard all enabled frameworks
                default = self.get_enabled_backends()
            else:
                default = [default]
        else:
            assert isinstance(default, list), "TODO"
        return default

    def get_default_frameworks(self):
        default = self.defaults.default_framework
        if default is None:
            return []
        if isinstance(default, str):
            if default == "*":  # Wildcard all enabled frameworks
                default = self.get_enabled_frameworks()
            else:
                default = [default]
        else:
            assert isinstance(default, list), "TODO"
        return default

    def get_default_targets(self):
        default = self.defaults.default_target
        if default is not None:
            if isinstance(default, str):
                if default == "*":  # Wildcard all enabled targets
                    default = self.get_enabled_targets()
                else:
                    default = [default]
            else:
                assert isinstance(default, list)
        return default


class DefaultEnvironment(Environment):
    def __init__(self):
        super().__init__()
        self.defaults = DefaultsConfig(
            log_level=logging.DEBUG,
            log_to_file=False,
            default_framework=None,
            default_backends={},
            default_target=None,
            cleanup_auto=False,
            cleanup_keep=100,
        )
        self.paths = {
            "deps": PathConfig("./deps"),
            "logs": PathConfig("./logs"),
            "results": PathConfig("./results"),
            "plugins": PathConfig("./plugins"),
            "temp": PathConfig("out"),
            "models": [
                PathConfig("./models"),
            ],
        }
        self.repos = {
            "tensorflow": RepoConfig("https://github.com/tensorflow/tensorflow.git", ref="v2.5.2"),
            "tflite_micro_compiler": RepoConfig(
                "https://github.com/cpetig/tflite_micro_compiler.git", ref="master"
            ),  # TODO: freeze ref?
            "tvm": RepoConfig(
                "https://github.com/tum-ei-eda/tvm.git", ref="tumeda"
            ),  # TODO: use upstream repo with suitable commit?
            "utvm_staticrt_codegen": RepoConfig(
                "https://github.com/tum-ei-eda/utvm_staticrt_codegen.git", ref="master"
            ),  # TODO: freeze ref?
            "etiss": RepoConfig("https://github.com/tum-ei-eda/etiss.git", ref="master"),  # TODO: freeze ref?
        }
        self.frameworks = [
            FrameworkConfig(
                "tflm",
                enabled=True,
                backends=[
                    BackendConfig("tflmc", enabled=True, features=[]),
                    BackendConfig("tflmi", enabled=True, features=[]),
                ],
                features=[
                    FrameworkFeatureConfig("muriscvnn", framework="tflm", supported=False),
                ],
            ),
            FrameworkConfig(
                "utvm",
                enabled=True,
                backends=[
                    BackendConfig(
                        "tvmaot",
                        enabled=True,
                        features=[
                            BackendFeatureConfig("unpacked_api", backend="tvmaot", supported=True),
                        ],
                    ),
                    BackendConfig("tvmrt", enabled=True, features=[]),
                    BackendConfig("tvmcg", enabled=True, features=[]),
                ],
                features=[
                    FrameworkFeatureConfig("memplan", framework="utvm", supported=False),
                ],
            ),
        ]
        self.frontends = [
            FrontendConfig("saved_model", enabled=False),
            FrontendConfig("ipynb", enabled=False),
            FrontendConfig(
                "tflite",
                enabled=True,
                features=[
                    FrontendFeatureConfig("packing", frontend="tflite", supported=False),
                ],
            ),
        ]
        self.vars = {
            "TEST": "abc",
        }
        self.flags = {}
        self.platforms = [
            PlatformConfig(
                "mlif",
                enabled=True,
                features=[PlatformFeatureConfig("debug", platform="mlif", supported=True)],
            )
        ]
        self.toolchains = {}
        self.targets = [
            TargetConfig(
                "etiss_pulpino",
                features=[
                    TargetFeatureConfig("debug", target="etiss_pulpino", supported=True),
                    TargetFeatureConfig("attach", target="etiss_pulpino", supported=True),
                    TargetFeatureConfig("trace", target="etiss_pulpino", supported=True),
                ],
            ),
            TargetConfig(
                "host_x86",
                features=[
                    TargetFeatureConfig("debug", target="host_x86", supported=True),
                    TargetFeatureConfig("attach", target="host_x86", supported=True),
                ],
            ),
        ]


class UserEnvironment(DefaultEnvironment):
    def __init__(
        self,
        home,
        merge=False,
        alias=None,
        defaults=None,
        paths=None,
        repos=None,
        frameworks=None,
        frontends=None,
        platforms=None,
        toolchains=None,
        targets=None,
        variables=None,
        default_flags=None,
    ):
        super().__init__()
        self._home = home

        if merge:
            raise NotImplementedError

        if alias:
            self.alias = alias
        if defaults:
            self.defaults = defaults
        if paths:
            self.paths = paths
        if repos:
            self.repos = repos
        if frameworks:
            self.frameworks = frameworks
        if frontends:
            self.frontends = frontends
        if platforms:
            self.platforms = platforms
        if toolchains:
            self.toolchains = toolchains
        if targets:
            self.targets = targets
        if variables:
            self.vars = variables
        if default_flags:
            self.flags = default_flags



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Command line subcommand for manaing environments."""

import configparser

from mlonmcu.environment.config import get_environments_file


def get_parser(subparsers):
    """ "Define and return a subparser for the env subcommand."""
    parser = subparsers.add_parser("env", description="List ML on MCU environments.")
    parser.set_defaults(func=handle)
    return parser


class EnvironmentHint:
    def __init__(self, name, path, created_at=None):
        self.name = name
        self.path = path
        self.created_at = created_at


def lookup_user_environments(file):
    config = configparser.ConfigParser()
    config.read(file)
    hints = []
    for env_name in config.sections():
        env = config[env_name]
        assert "path" in env
        path = env["path"]
        created_at = None
        if "created" in env:
            created_at = env["created"]
        hint = EnvironmentHint(env_name, path, created_at=created_at)
        hints.append(hint)
    return hints


def handle(args):
    envs_file = get_environments_file()
    print(f"Looking for user environments config file: {envs_file}")
    envs = lookup_user_environments(envs_file)
    count = len(envs)
    if not envs or count == 0:
        print("No mlonmcu environments were found for the current user. You can create a new one using `mlonmcu init`")
    else:
        print(f"Found {count} mlonmcu environment(s):")
        for env in envs:
            print(f"    - {env.name} ({env.path})")

        print("Point the environment variable MLONMCU_HOME to a environment directory to get started")



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import re
import multiprocessing
from pathlib import Path

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import _validate_gcc, get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def check_multilibs(riscvInstallDir, gccName, live=False, vext=False, pext=False):
    gccExe = Path(riscvInstallDir) / "bin" / f"{gccName}-gcc"
    out = utils.exec_getout(gccExe, "--print-multi-lib", live=live)
    multilibs = []
    if vext and pext:
        default_multilib = "rv32gcpv/ilp32d" if "32" in gccName else "rv64gcpv/lp64d"  # TODO: improve this
    elif vext:
        default_multilib = "rv32gcv/ilp32d" if "32" in gccName else "rv64gcv/lp64d"  # TODO: improve this
    elif pext:
        default_multilib = "rv32gcp/ilp32d" if "32" in gccName else "rv64gcp/lp64d"  # TODO: improve this
    else:
        default_multilib = "rv32gc/ilp32d" if "32" in gccName else "rv64gc/lp64d"  # TODO: improve this
    if len(out.split("\n")) < 3:
        multilib = False
    else:
        multilib = True
        libs = re.compile(r".+\/.+;@march=(.+)@mabi=(.+)").findall(out)
        for arch, abi in libs:
            multilibs.append(f"{arch}/{abi}")

    return multilib, default_multilib, multilibs


def _validate_riscv_gcc(context: MlonMcuContext, params=None):
    if not _validate_gcc(context, params=params):
        return False
    if not (
        context.environment.has_target("etiss_pulpino")
        or context.environment.has_target("etiss")
        or context.environment.has_target("spike")
        or context.environment.has_target("ovpsim")
        or context.environment.has_target("ara")
    ):
        return False
    if params:
        vext = params.get("vext", False)
        pext = params.get("pext", False)
        user_vars = context.environment.vars
        combined = user_vars.get("riscv_gcc.combined", False)
        if vext and pext:
            return combined
        elif vext:
            if not context.environment.has_feature("vext"):
                return False
        elif pext:
            if not context.environment.has_feature("pext"):
                return False
    return True


@Tasks.provides(
    [
        "riscv_gcc.install_dir",
        "riscv_gcc.name",
        "riscv_gcc.variant",
        "riscv_gcc.multilib",
        "riscv_gcc.default_multilib",
        "riscv_gcc.multilibs",
    ]
)
@Tasks.param("vext", [False, True])
@Tasks.param("pext", [False, True])
@Tasks.validate(_validate_riscv_gcc)
@Tasks.register(category=TaskType.TOOLCHAIN)
def install_riscv_gcc(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download and install the RISCV GCC toolchain."""
    if not params:
        params = {}
    user_vars = context.environment.vars
    vext = params["vext"]
    pext = params["pext"]
    combined = user_vars.get("riscv_gcc.combined", False)
    variant = user_vars.get("riscv_gcc.variant", "unknown")
    flags = utils.makeFlags((params["vext"], "vext"), (params["pext"], "pext"))
    # TODO: if the used gcc supports both pext and vext we do not need to download it 3 times!
    if combined:
        riscvName = utils.makeDirName("riscv_gcc", flags=[])
    else:
        riscvName = utils.makeDirName("riscv_gcc", flags=flags)
    riscvInstallDir = context.environment.paths["deps"].path / "install" / riscvName
    if (not vext) and (not pext) and "riscv_gcc.install_dir_default" in user_vars:
        riscvInstallDir = Path(user_vars["riscv_gcc.install_dir_default"])
        multilib = user_vars.get("riscv_gcc.multilib_default", None)
        default_multilib = user_vars.get("riscv_gcc_default.default_multilib_default", None)
        multilibs = user_vars.get("riscv_gcc.multilibs_default", None)
    elif vext and "riscv_gcc.install_dir_vext" in user_vars:
        riscvInstallDir = Path(user_vars["riscv_gcc.install_dir_vext"])
        multilib = user_vars.get("riscv_gcc.multilib_vext", None)
        default_multilib = user_vars.get("riscv_gcc.default_multilib_vext", None)
        multilibs = user_vars.get("riscv_gcc.multilibs_vext", None)
    elif pext and "riscv_gcc.install_dir_pext" in user_vars:
        riscvInstallDir = Path(user_vars["riscv_gcc.install_dir_pext"])
        multilib = user_vars.get("riscv_gcc.multilib_pext", None)
        default_multilib = user_vars.get("riscv_gcc.default_multilib_pext", None)
        multilibs = user_vars.get("riscv_gcc.multilibs_pext", None)
    elif "riscv_gcc.install_dir" in user_vars:  # TODO: also check command line flags?
        # This would overwrite the cache.ini entry which is NOT wanted! -> return false but populate gcc_name?
        riscvInstallDir = Path(user_vars["riscv_gcc.install_dir"])
        multilib = user_vars.get("riscv_gcc.multilib", None)
        default_multilib = user_vars.get("riscv_gcc.default_multilib", None)
        multilibs = user_vars.get("riscv_gcc.multilibs", None)
    else:

        def _helper(url):
            fullUrlSplit = url.split("/")
            riscvUrl = "/".join(fullUrlSplit[:-1])
            riscvFileName, riscvFileExtension = fullUrlSplit[-1].split(".", 1)
            return riscvUrl, riscvFileName, riscvFileExtension

        if vext and "riscv_gcc.dl_url_vext" in user_vars:
            assert not combined, "Combined toolchain does only support riscv_gcc.dl_url"
            riscvUrl, riscvFileName, riscvFileExtension = _helper(user_vars["riscv_gcc.dl_url_vext"])
        elif pext and "riscv_gcc.dl_url_pext" in user_vars:
            assert not combined, "Combined toolchain does only support riscv_gcc.dl_url"
            riscvUrl, riscvFileName, riscvFileExtension = _helper(user_vars["riscv_gcc.dl_url_pext"])
        elif "riscv_gcc.dl_url" in user_vars:
            riscvUrl, riscvFileName, riscvFileExtension = _helper(user_vars["riscv_gcc.dl_url"])
        else:
            riscvVersion = (
                user_vars["riscv.version"]
                if "riscv.version" in user_vars
                else ("8.3.0-2020.04.0" if not vext else "10.2.0-2020.12.8")
            )
            riscvDist = (
                user_vars["riscv.distribution"] if "riscv.distribution" in user_vars else "x86_64-linux-ubuntu14"
            )
            if vext:
                subdir = "v" + ".".join(riscvVersion.split("-")[1].split(".")[:-1])
                riscvUrl = "https://static.dev.sifive.com/dev-tools/freedom-tools/" + subdir + "/"
                riscvFileName = f"riscv64-unknown-elf-toolchain-{riscvVersion}-{riscvDist}"
            else:
                riscvUrl = "https://static.dev.sifive.com/dev-tools/"
                riscvFileName = f"riscv64-unknown-elf-gcc-{riscvVersion}-{riscvDist}"
            riscvFileExtension = "tar.gz"
        riscvArchive = riscvFileName + "." + riscvFileExtension
        # if rebuild or not utils.is_populated(riscvInstallDir):
        # rebuild should only be triggered if the version/url changes but we can not detect that at the moment
        if not utils.is_populated(riscvInstallDir):
            utils.download_and_extract(riscvUrl, riscvArchive, riscvInstallDir, progress=verbose)
        multilib = user_vars.get("riscv_gcc.multilib", None)
        default_multilib = user_vars.get("riscv_gcc.default_multilib", None)
        multilibs = user_vars.get("riscv_gcc.multilibs", None)
    assert utils.is_populated(riscvInstallDir)
    if "riscv_gcc.name" in user_vars:
        gccName = user_vars["riscv_gcc.name"]
    else:
        gccNames = ["riscv64-unknown-elf", "riscv32-unknown-elf"]
        gccName = None
        for name in gccNames:
            if (Path(riscvInstallDir) / name).is_dir():
                gccName = name
                break
    assert gccName is not None, "Toolchain name could not be determined automatically"
    multilib_, default_multilib_, multilibs_ = check_multilibs(
        riscvInstallDir, gccName, live=verbose, vext=vext, pext=pext
    )
    context.cache["riscv_gcc.install_dir", flags] = riscvInstallDir
    context.cache["riscv_gcc.name", flags] = gccName
    context.cache["riscv_gcc.variant", flags] = variant
    context.cache["riscv_gcc.multilib", flags] = multilib or multilib_
    context.cache["riscv_gcc.default_multilib", flags] = default_multilib or default_multilib_
    context.cache["riscv_gcc.multilibs", flags] = multilibs or multilibs_
    context.export_paths.add(riscvInstallDir / "bin")



# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for the tvm wrapper generator."""
import pytest

from mlonmcu.flow.tvm.backend import wrapper
from mlonmcu.flow.tvm.backend.model_info import TensorInfo, ModelInfo

DUMMY_GRAPH_JSON = "{}"
DUMMY_PARAMS = bytes()

MODEL_INFO_1 = ModelInfo(
    [
        TensorInfo("input", (1024,), "int8"),
    ],
    [
        TensorInfo("output", (8,), "int8"),
    ],
)
MODEL_INFO_2 = ModelInfo(
    [
        TensorInfo("input", (10, 10), "float32"),
        TensorInfo("input2", (1, 16), "float32"),
    ],
    [
        TensorInfo("output", (8,), "float32"),
        TensorInfo("output2", (2, 4), "float32"),
    ],
)


def _check(out, expected_lines):
    out_lines = out.split("\n")[2:]
    for expected_line in expected_lines:
        assert expected_line in out_lines


def test_calc_pages():
    # valid
    assert wrapper.calc_pages(2**15, page_size=2**10) == (32, 10)
    assert wrapper.calc_pages((2**15) - 1, page_size=2**10) == (32, 10)
    assert wrapper.calc_pages(2**15, page_size=2**11) == (16, 11)
    # invalid
    with pytest.raises(AssertionError):
        wrapper.calc_pages(-1)
    with pytest.raises(AssertionError):
        wrapper.calc_pages(2**15, page_size=123)


@pytest.mark.parametrize("model_info", [MODEL_INFO_1, MODEL_INFO_2])
@pytest.mark.parametrize("workspace_size", [2**15])
@pytest.mark.parametrize("debug_arena", [False, True])
def test_wrapper_graph(model_info, workspace_size, debug_arena):
    out = wrapper.generate_tvmrt_wrapper(DUMMY_GRAPH_JSON, DUMMY_PARAMS, model_info, workspace_size, debug_arena)
    expected_lines = [
        "#include <stdlib.h>",
        "#include <stdarg.h>",
        "#include <dlpack/dlpack.h>",
        '#include "tvm/runtime/c_runtime_api.h"',
        '#include "tvm/runtime/crt/error_codes.h"',
        '#include "printing.h"',
    ]
    expected_lines.extend(
        [
            '#include "tvm/runtime/crt/packed_func.h"',
            '#include "tvm/runtime/crt/crt.h"',
            '#include "tvm/runtime/crt/graph_executor.h"',
            '#include "tvm/runtime/crt/page_allocator.h"',
        ]
    )
    # expected_lines.extend(
    #     [
    #         "char input0_data[400];",
    #         "char input1_data[64];",
    #         "char output0_data[32];",
    #         "char output1_data[32];",
    #         "void* inputs[] = {input0_data, input1_data};",
    #         "void* outputs[] = {output0_data, output1_data};",
    #     ]
    # )
    expected_lines.extend(
        [
            "tvm_crt_error_t TVMPlatformMemoryAllocate(size_t num_bytes, DLDevice dev, void** out_ptr)",
            "    tvm_crt_error_t ret = g_memory_manager->Allocate(g_memory_manager, num_bytes, dev, out_ptr);",
            "tvm_crt_error_t TVMPlatformMemoryFree(void* ptr, DLDevice dev)",
            "    return g_memory_manager->Free(g_memory_manager, ptr, dev);",
            "    TVM_CCALL(PageMemoryManagerCreate(&g_memory_manager, g_crt_memory, sizeof(g_crt_memory),",
            "                                      CRT_MEMORY_PAGE_SIZE_LOG2));",
            "    TVM_CCALL(TVMInitializeRuntime());",
            '    TVM_CCALL(TVMPackedFunc_InitGlobalFunc(&pf, "runtime.SystemLib", &args));',
            "    TVM_CCALL(TVMPackedFunc_Call(&pf));",
            "    TVM_CCALL(TVMGraphExecutor_Create(g_graph, mod_syslib, &dev, &graph_executor));",
            "    TVM_CCALL(TVMGraphExecutor_LoadParams(graph_executor, params.data, params.size));",
            "#define CRT_MEMORY_NUM_PAGES 32",
            "#define CRT_MEMORY_PAGE_SIZE_LOG2 10",
            "static uint8_t g_crt_memory[CRT_MEMORY_NUM_PAGES * (1 << CRT_MEMORY_PAGE_SIZE_LOG2)];",
            "static MemoryManagerInterface* g_memory_manager;",
        ]
    )
    expected_lines.extend(
        [
            "void __attribute__((noreturn)) TVMPlatformAbort(tvm_crt_error_t code)",
            "int TVMWrap_Init()",
        ]
    )
    expected_lines.extend(
        [
            "void *TVMWrap_GetInputPtr(int index)",
            # "    return inputs[index];",
            "size_t TVMWrap_GetInputSize(int index)",
            # "    size_t sizes[] = { 400, 64, };",
            # "    return sizes[index];",
            "size_t TVMWrap_GetNumInputs()",
            # "    return 2;",
        ]
    )
    expected_lines.extend(
        [
            "int TVMWrap_Run()",
        ]
    )
    expected_lines.extend(
        [
            # "    if (ret_val)",
            # "        TVMPlatformAbort(kTvmErrorPlatformCheckFailure);",
        ]
    )
    if debug_arena and workspace_size > 0:
        expected_lines.extend(
            [
                "#if DEBUG_ARENA_USAGE",
                (
                    '    DBGPRINTF("\\nGraph executor arena max usage after model invocation: %lu bytes\\n",'
                    " max_arena_usage);"
                ),
                "#endif  // DEBUG_ARENA_USAGE",
            ]
        )
    expected_lines.extend(
        [
            "void *TVMWrap_GetOutputPtr(int index)",
            # "    return outputs[index];",
            "size_t TVMWrap_GetOutputSize(int index)",
            # "    size_t sizes[] = { 32, 32, };",
            # "    return sizes[index];",
            "size_t TVMWrap_GetNumOutputs()",
            # "    return 2;",
        ]
    )
    _check(out, expected_lines)


@pytest.mark.parametrize("model_info", [MODEL_INFO_1, MODEL_INFO_2])
@pytest.mark.parametrize("workspace_size", [0, 2**15])
@pytest.mark.parametrize("mod_name", ["default", "my_module"])
@pytest.mark.parametrize("api", ["c", "packed"])
@pytest.mark.parametrize("debug_arena", [False, True])
def test_wrapper_aot(model_info, workspace_size, mod_name, api, debug_arena):
    out = wrapper.generate_tvmaot_wrapper(model_info, workspace_size, mod_name, api, debug_arena)
    expected_lines = [
        "#include <stdlib.h>",
        "#include <stdarg.h>",
        "#include <dlpack/dlpack.h>",
        '#include "tvm/runtime/c_runtime_api.h"',
        '#include "tvm/runtime/crt/error_codes.h"',
        '#include "printing.h"',
    ]
    if workspace_size != 0:
        expected_lines.append('#include "tvm/runtime/crt/stack_allocator.h"')
    # expected_lines.extend(
    #     [
    #         "char input0_data[400];",
    #         "char input1_data[64];",
    #         "char output0_data[32];",
    #         "char output1_data[32];",
    #         "void* inputs[] = {input0_data, input1_data};",
    #         "void* outputs[] = {output0_data, output1_data};",
    #     ]
    # )
    if api == "packed":
        pass
        # expected_lines.extend(
        #     [
        #         "void* inputs[] = {input0_data, input1_data};",
        #         "void* outputs[] = {output0_data, output1_data};",
        #     ]
        # )
    elif api == "packed":
        expected_lines.extend(
            [
                f'#include "tvmgen_{mod_name}.h"',
                f"struct tvmgen_{mod_name}_inputs tvmgen_{mod_name}_inputs = {{",
                f"struct tvmgen_{mod_name}_outputs tvmgen_{mod_name}_outputs = {{",
            ]
        )
    if workspace_size > 0:
        expected_lines.extend(
            [
                f"#define WORKSPACE_SIZE ({int(workspace_size)})",
                "static uint8_t g_aot_memory[WORKSPACE_SIZE];",
                "tvm_workspace_t app_workspace;",
                "tvm_crt_error_t TVMPlatformMemoryAllocate(size_t num_bytes, DLDevice dev, void** out_ptr)",
                "    tvm_crt_error_t ret = StackMemoryManager_Allocate(&app_workspace, num_bytes, out_ptr);",
                "tvm_crt_error_t TVMPlatformMemoryFree(void* ptr, DLDevice dev)",
                "    return StackMemoryManager_Free(&app_workspace, ptr);",
                "    StackMemoryManager_Init(&app_workspace, g_aot_memory, WORKSPACE_SIZE);",
            ]
        )
    expected_lines.extend(
        [
            "void __attribute__((noreturn)) TVMPlatformAbort(tvm_crt_error_t code)",
            "TVM_DLL int TVMFuncRegisterGlobal(const char* name, TVMFunctionHandle f, int override)",
            "int TVMWrap_Init()",
        ]
    )
    expected_lines.extend(
        [
            "void *TVMWrap_GetInputPtr(int index)",
            "    return inputs[index];",
            "size_t TVMWrap_GetInputSize(int index)",
            # "    size_t sizes[] = { 400, 64, };",
            "    return sizes[index];",
            "size_t TVMWrap_GetNumInputs()",
            # "    return 2;",
        ]
    )
    if api == "packed":
        expected_lines.extend(
            [
                (
                    f"int32_t tvmgen_{mod_name}_run(void* args, void* type_code, int num_args, void* out_value, void*"
                    " out_type_code, void* resource_handle);"
                ),
                "int TVMWrap_Run()",
                "    static DLDevice fake_device = {kDLCPU, 0};",
                "    static int64_t fake_dims = 0;",
                "    static int64_t fake_shape = {0};",
                # "    DLTensor tensors[2 + 2];",
                # "    TVMValue values[2 + 2];",
                # "    int32_t typeids[2 + 2];",
                # "    for (size_t i = 0; i < 2+2; i++) {",
                "        tensors[i].device = fake_device;",
                # "        tensors[i].data = (i < 2) ? inputs[i] : outputs[i - 2];",
                "        tensors[i].shape = &fake_shape;",
                "        tensors[i].ndim = fake_dims;",
                "        tensors[i].byte_offset = 0;",
                "        tensors[i].strides = NULL;",
                "        values[i].v_handle = &tensors[i];",
                f"    int ret_val = tvmgen_{mod_name}_run(values, typeids, 0, NULL, 0, NULL);",
            ]
        )
    elif api == "c":
        expected_lines.extend(
            [
                "int TVMWrap_Run()",
                f"    int ret_val = tvmgen_{mod_name}_run(&tvmgen_{mod_name}_inputs, &tvmgen_{mod_name}_outputs);",
            ]
        )
    expected_lines.extend(
        [
            "    if (ret_val)",
            "        TVMPlatformAbort(kTvmErrorPlatformCheckFailure);",
        ]
    )
    if debug_arena and workspace_size > 0:
        expected_lines.extend(
            [
                "#if DEBUG_ARENA_USAGE",
                (
                    '    DBGPRINTF("\\nAoT executor arena max usage after model invocation: %lu bytes\\n",'
                    " max_arena_usage);"
                ),
                "#endif  // DEBUG_ARENA_USAGE",
            ]
        )
    expected_lines.extend(
        [
            "void *TVMWrap_GetOutputPtr(int index)",
            "    return outputs[index];",
            "size_t TVMWrap_GetOutputSize(int index)",
            # "    size_t sizes[] = { 32, 32, };",
            "    return sizes[index];",
            "size_t TVMWrap_GetNumOutputs()",
            # "    return 2;",
        ]
    )
    _check(out, expected_lines)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
from math import sqrt
from enum import Enum
from pathlib import Path
from collections import namedtuple

from mlonmcu.config import filter_config

from .metadata import parse_metadata

ModelFormat = namedtuple("ModelFormat", ["value", "extensions"])


class ModelFormats(Enum):
    @property
    def extensions(self):
        return self.value.extensions

    @property
    def extension(self):
        return self.value.extensions[0]

    @classmethod
    def from_extension(cls, ext):
        for fmt in cls:
            if ext in fmt.extensions:
                return fmt
        return None

    NONE = ModelFormat(0, [])
    TFLITE = ModelFormat(1, ["tflite"])
    PACKED = ModelFormat(2, ["tflm"])
    IPYNB = ModelFormat(3, ["ipynb"])
    ONNX = ModelFormat(4, ["onnx"])
    RELAY = ModelFormat(5, ["relay"])
    PB = ModelFormat(6, ["pb"])
    PADDLE = ModelFormat(7, ["pdmodel"])
    TEXT = ModelFormat(8, ["txt"])


def parse_metadata_from_path(path):
    if Path(path).is_file():
        metadata = parse_metadata(path)
        return metadata
    return None


def parse_shape_string(inputs_string):
    """Parse an input shape dictionary string to a usable dictionary.

    Taken from: https://github.com/apache/tvm/blob/main/python/tvm/driver/tvmc/shape_parser.py

    Parameters
    ----------
    inputs_string: str
        A string of the form "input_name:[dim1,dim2,...,dimn] input_name2:[dim1,dim2]" that
        indicates the desired shape for specific model inputs. Colons, forward slashes and dots
        within input_names are supported. Spaces are supported inside of dimension arrays.
    Returns
    -------
    shape_dict: dict
        A dictionary mapping input names to their shape for use in relay frontend converters.
    """

    # Create a regex pattern that extracts each separate input mapping.
    # We want to be able to handle:
    # * Spaces inside arrays
    # * forward slashes inside names (but not at the beginning or end)
    # * colons inside names (but not at the beginning or end)
    # * dots inside names
    pattern = r"(?:\w+\/)?[:\w.]+\:\s*\[\-?\d+(?:\,\s*\-?\d+)*\]"
    input_mappings = re.findall(pattern, inputs_string)
    assert input_mappings, f"Invalid shapes string: {inputs_string} (Expected syntax: 'foo:[1,32,32,3] bar:[10,10]')"
    shape_dict = {}
    for mapping in input_mappings:
        # Remove whitespace.
        mapping = mapping.replace(" ", "")
        # Split mapping into name and shape.
        name, shape_string = mapping.rsplit(":", 1)
        # Convert shape string into a list of integers or Anys if negative.
        shape = [int(x) for x in shape_string.strip("][").split(",")]
        # Add parsed mapping to shape dictionary.
        shape_dict[name] = shape

    return shape_dict


def parse_type_string(inputs_string):
    """Parse an input type dictionary string to a usable dictionary.

    Parameters
    ----------
    inputs_string: str
        A string of the form "input_name:ty input_name2:ty" that
        indicates the desired type for specific model inputs/outputs. Colons, forward slashes and dots
        within input_names are supported. Spaces are supported inside of dimension arrays.
    Returns
    -------
    type_dict: dict
        A dictionary mapping input names to their type.
    """
    pattern = r"(?:\w+\/)?[:\w.]+\:\s*\-?\w+(?:\s*\-?\w+)*"
    input_mappings = re.findall(pattern, inputs_string)
    assert input_mappings, f"Invalid types string: {inputs_string} (Expected syntax: 'foo:int8 bar:int32')"
    type_dict = {}
    for mapping in input_mappings:
        # Remove whitespace.
        mapping = mapping.replace(" ", "")
        # Split mapping into name and type.
        name, type_string = mapping.rsplit(":", 1)
        type_dict[name] = type_string

    return type_dict


class Workload:
    DEFAULTS = {}

    def __init__(self, name, config=None, alt=None):
        self.name = name
        self.alt = alt
        self.config = filter_config(config if config is not None else {}, self.name, self.DEFAULTS, set(), set())

    def get_platform_config(self, platform):
        return {}

    def add_platform_config(self, platform, config):
        config.update(self.get_platform_config(platform))

    def get_platform_defs(self, platform):
        return {}

    def add_platform_defs(self, platform, defs):
        defs.update(self.get_platform_defs(platform))


class Model(Workload):
    DEFAULTS = {
        **Workload.DEFAULTS,
        "metadata_path": "definition.yml",
        "input_shapes": None,
        "output_shapes": None,
        "input_types": None,
        "output_types": None,
        "support_path": "support",
        "inputs_path": "input",
        "outputs_path": "output",
    }

    def __init__(self, name, paths, config=None, alt=None, formats=ModelFormats.TFLITE):
        super().__init__(name, config=config, alt=alt)
        self.paths = paths
        if not isinstance(self.paths, list):
            self.paths = [self.path]
        self.formats = formats
        if not isinstance(self.formats, list):
            self.formats = [formats]
        self.metadata = parse_metadata_from_path(self.metadata_path)

    @property
    def metadata_path(self):
        return self.config["metadata_path"]

    @property
    def input_shapes(self):
        temp = self.config["input_shapes"]
        if temp:
            if isinstance(temp, str):
                temp = parse_shape_string(temp)
            else:
                assert isinstance(temp, dict)
        return temp

    @property
    def output_shapes(self):
        temp = self.config["output_shapes"]
        if temp:
            if isinstance(temp, str):
                temp = parse_shape_string(temp)
            else:
                assert isinstance(temp, dict)
        return temp

    @property
    def input_types(self):
        temp = self.config["input_types"]
        if temp:
            if isinstance(temp, str):
                temp = parse_type_string(temp)
            else:
                assert isinstance(temp, dict)
        return temp

    @property
    def output_types(self):
        temp = self.config["output_types"]
        if temp:
            if isinstance(temp, str):
                temp = parse_type_string(temp)
            else:
                assert isinstance(temp, dict)
        return temp

    @property
    def support_path(self):
        return self.config["support_path"]

    @property
    def inputs_path(self):
        # TODO: fall back to metadata
        return self.config["inputs_path"]

    @property
    def outputs_path(self):
        # TODO: fall back to metadata
        return self.config["outputs_path"]

    @property
    def skip_check(self):
        if len(self.formats) == 0:
            return True
        elif len(self.formats) == 1:
            return self.formats[0] == ModelFormats.TEXT
        else:
            return False

    def __repr__(self):
        if self.alt:
            return f"Model({self.name},alt={self.alt})"
        return f"Model({self.name})"


class Program(Workload):
    def __repr__(self):
        if self.alt:
            return f"Program({self.name},alt={self.alt})"
        return f"Program({self.name})"


class ExampleProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["EXAMPLE_BENCHMARK"] = self.name
        return ret


class EmbenchProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["EMBENCH_BENCHMARK"] = self.name
        return ret


class TaclebenchProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["TACLEBENCH_BENCHMARK"] = self.name
        return ret


class PolybenchProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["POLYBENCH_BENCHMARK"] = self.name
        return ret


class MibenchProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["MIBENCH_BENCHMARK"] = self.name
        return ret


class MathisProgram(Program):
    DEFAULTS = {
        "size": 1024,
        # "size": 65536,
    }

    @property
    def size(self):
        value = self.config["size"]
        if isinstance(value, str):
            value = str(value)
        assert isinstance(value, int)
        assert value > 0
        return value

    def get_nargs(self, name):
        return {
            "to_upper": 2,
            "add8": 4,
            "add16": 4,
            "gather_add8": 4,
            "gather_add16": 4,
            "scatter_add8": 4,
            "scatter_add16": 4,
            "dot8": 3,
            "dot16": 3,
            "saxpy8": 5,
            "saxpy16": 5,
            "matmul8": 4,
            "matmul16": 4,
            "matmul8_a": 4,
            "matmul16_a": 4,
            "transposed_matmul8": 4,
            "transposed_matmul16": 4,
            "transposed_matmul8_a": 4,
            "transposed_matmul16_a": 4,
            "transposed_matmul8_b": 4,
            "transposed_matmul16_b": 4,
        }[name]

    def get_elem_size(self, name):
        return {
            "to_upper": 8,
            "add8": 8,
            "add16": 16,
            "gather_add8": 8,
            "gather_add16": 16,
            "scatter_add8": 8,
            "scatter_add16": 16,
            "dot8": 8,
            "dot16": 16,
            "saxpy8": 8,
            "saxpy16": 16,
            "matmul8": 8,
            "matmul16": 16,
            "matmul8": 8,
            "matmul16": 16,
            "matmul8_a": 8,
            "matmul16_a": 16,
            "transposed_matmul8": 8,
            "transposed_matmul16": 16,
            "transposed_matmul8_a": 8,
            "transposed_matmul16_a": 16,
            "transposed_matmul8_b": 8,
            "transposed_matmul16_b": 16,
        }[name]

    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["MATHIS_TEST"] = self.name
            ret["MATHIS_NARGS"] = self.get_nargs(self.name)
            ret["MATHIS_ELEM_SIZE"] = self.get_elem_size(self.name)
            ret["MATHIS_SIZE"] = self.size
            ret["MATHIS_N"] = int(sqrt(self.size)) if "matmul" in self.name else self.size
        return ret


class CoremarkProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["COREMARK_ITERATIONS"] = 10
        return ret


class DhrystoneProgram(Program):
    def get_platform_defs(self, platform):
        ret = {}
        if platform == "mlif":
            ret["DHRYSTONE_ITERATIONS"] = 10000
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU postprocess submodule"""

from .postprocesses import (
    FilterColumnsPostprocess,
    RenameColumnsPostprocess,
    Features2ColumnsPostprocess,
    Config2ColumnsPostprocess,
    MyPostprocess,
    PassConfig2ColumnsPostprocess,
    VisualizePostprocess,
    Bytes2kBPostprocess,
    Artifact2ColumnPostprocess,
    AnalyseInstructionsPostprocess,
    CompareRowsPostprocess,
    AnalyseDumpPostprocess,
    AnalyseCoreVCountsPostprocess,
)

SUPPORTED_POSTPROCESSES = {
    "filter_cols": FilterColumnsPostprocess,
    "rename_cols": RenameColumnsPostprocess,
    "features2cols": Features2ColumnsPostprocess,
    "config2cols": Config2ColumnsPostprocess,
    "mypost": MyPostprocess,
    "passcfg2cols": PassConfig2ColumnsPostprocess,
    "visualize": VisualizePostprocess,
    "bytes2kb": Bytes2kBPostprocess,
    "artifacts2cols": Artifact2ColumnPostprocess,
    "analyse_instructions": AnalyseInstructionsPostprocess,
    "compare_rows": CompareRowsPostprocess,
    "analyse_dump": AnalyseDumpPostprocess,
    "analyse_corev_counts": AnalyseCoreVCountsPostprocess,
}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Transform passes for the q_vanilla_accelerator accelerator"""

# import functools
import tvm
from tvm import tir

# from tvm.relay.backend.contrib.uma.api.utils import add_llvm_to_block
from tvm import relay

# from tvm.tir import buffer


@tvm.tir.transform.prim_func_pass(opt_level=2)
class QVanillaAcceleratorConv2dPass:
    _EXTERNAL_FUNCTION_NAME = "q_vanilla_accelerator_conv2dnchw"
    # _TVM_BLOCK_MATCH_NAME = "conv2d_nchw"
    _TVM_BLOCK_MATCH_NAME = "compute_2"
    print("tir pass")

    def transform_function(
        self, func: tvm.tir.PrimFunc, mod: tvm.ir.IRModule, ctx: tvm.ir.transform.PassContext
    ) -> tvm.tir.PrimFunc:
        return self._q_vanilla_accelerator_conv2d_pass(func, mod, ctx)

    @classmethod
    def _q_vanilla_accelerator_conv2d_pass(cls, func, mod, ctx):
        _loops = dict()  # noqa: F841
        _handles = []  # noqa: F841
        _entry_node = None  # noqa: F841
        zp = []
        block_idx = 0  # noqa: F841

        def _has_block(name: str, func: tvm.tir.PrimFunc) -> bool:
            """
            Determine of a tir.block with `name` exists in `func`
            """

            def _hb(op):
                if isinstance(op, tvm.tir.Block):
                    _found_blocks.append(op.name_hint)

            _found_blocks = []
            tvm.tir.stmt_functor.post_order_visit(func.body, _hb)

            return name in _found_blocks

        def _detect_and_replace_conv2d(
            func: tvm.tir.PrimFunc, mod: tvm.ir.IRModule, ctx: tvm.ir.transform.PassContext
        ) -> tvm.tir.PrimFunc:
            def _replace_conv2d(op):
                if op == _entry_node:
                    print("if op == _entry_node")
                    irb = tvm.tir.ir_builder.create()
                    # Collection of buffer address
                    buffers = [b[1].data for b in _handles]
                    # extraction of loop offsets
                    for k, v in _loops.items():
                        assert v.min.value == 0
                    offset_order = ["co", "w", "h", "ci", "kh", "kw"]
                    offsets = [_loops[i].extent.value for i in offset_order]

                    offsets.append(zp[0])
                    offsets.append(zp[1])

                    args = buffers + offsets

                    irb.emit(tir_call(irb, True, cls._EXTERNAL_FUNCTION_NAME, *args))
                    irb_result = irb.get()

                    return irb_result
                elif isinstance(op, tvm.tir.SeqStmt):
                    # Remove that pad block of TOPI's conv2DNCHW by only returning the 2nd statement

                    return op.seq[block_idx]  # the line that I've changed to replace the compute_2 block

                return op

            sch = tir.Schedule(func)

            if _has_block(cls._TVM_BLOCK_MATCH_NAME, func):
                # find the zp values

                s1 = []
                s2 = []

                def _visit(s):
                    if isinstance(s, tvm.tir.BufferStore):
                        # stores.append(s.value)
                        s1.append(s.buffer.data)
                        s2.append(s.value)

                tvm.tir.stmt_functor.post_order_visit(func.body, _visit)

                for i in range(len(s1)):
                    if s1[i].name == "compile_engine_const":
                        zp.append(s2[i])
                block_idx = len(s1) - 3

                ###

                conv2d_block = sch.get_block(cls._TVM_BLOCK_MATCH_NAME)
                rv_loops = sch.get_loops(conv2d_block)

                assert len(rv_loops) == 7
                loops = dict(
                    n=rv_loops[0],
                    co=rv_loops[1],
                    h=rv_loops[2],
                    w=rv_loops[3],
                    ci=rv_loops[4],
                    kh=rv_loops[5],
                    kw=rv_loops[6],
                )
                _entry_node = sch.get(rv_loops[1])

                _loops = {k: sch.get(v) for k, v in loops.items()}
                _handles = func.buffer_map.items()

                x = tvm.tir.stmt_functor.ir_transform(func.body, None, _replace_conv2d, ["tir.For", "tir.SeqStmt"])

                return func.with_body(x)
            else:
                # print(func)
                return func

        r = _detect_and_replace_conv2d(func, mod, ctx)
        return r


def tir_call(ib: tvm.tir.ir_builder, extern: bool, name: str, *args):
    """
    ib: ir_builder
    extern: bool
        True  --> tvm.tir.call_extern
        False --> tvm.tir.call_packed
    name: str
        function name
    *args:
        arguments for function call
    """

    def buf_from_array(ib, arr, dtype):
        # Allocate enough memory to store the whole array
        var = ib.allocate("int32", (len(arr),), scope="global")
        for i, v in enumerate(arr):
            var[i] = v
        # Declare a buffer, which is basically a view on the chunk of memory that we allocated
        buf = tvm.tir.decl_buffer((len(arr),), dtype, data=var, scope="global")
        return buf

    if extern:
        args = [i.data if isinstance(i, tvm.tir.Buffer) else i for i in args]
        return tvm.tir.call_extern("int32", name, *args)
    else:
        args = [
            buf_from_array(ib, i, "int32") if isinstance(i, (tuple, list, tvm.ir.container.Array)) else i for i in args
        ]
        return tvm.tir.call_packed(name, *args)


@tvm.ir.transform.module_pass(opt_level=0)
class ConvertLayout:
    def transform_module(self, mod, ctx):
        # My pass functionality...
        desired_layouts = {"qnn.conv2d": ["NCHW", "default"]}
        # Convert the layout to NCHW
        # RemoveUnunsedFunctions is used to clean up the graph.
        seq = tvm.transform.Sequential(
            [relay.transform.RemoveUnusedFunctions(), relay.transform.ConvertLayout(desired_layouts)]
        )
        with tvm.transform.PassContext(opt_level=3):
            mod = seq(mod)

        return mod


@tvm.ir.transform.module_pass(opt_level=0)
class Canonicalize:
    print("Canonicalize")

    def transform_module(self, mod, ctx):
        # My pass functionality...

        # Convert the layout to NCHW
        # RemoveUnunsedFunctions is used to clean up the graph.
        seq = tvm.transform.Sequential([relay.transform.RemoveUnusedFunctions(), relay.qnn.transform.CanonicalizeOps()])
        with tvm.transform.PassContext(opt_level=3):
            mod = seq(mod)

        return mod



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definition of tasks used to dynamically install MLonMCU dependencies"""

import multiprocessing

from mlonmcu.setup.task import TaskType
from mlonmcu.context.context import MlonMcuContext
from mlonmcu.setup import utils
from mlonmcu.logging import get_logger

from .common import get_task_factory

logger = get_logger()

Tasks = get_task_factory()


def _validate_tflite_visualize(context: MlonMcuContext, params=None):
    return context.environment.has_frontend("tflite") and context.environment.has_feature("visualize")


@Tasks.provides(["tflite_visualize.exe"])
@Tasks.validate(_validate_tflite_visualize)
@Tasks.register(category=TaskType.FEATURE)
def download_tflite_vizualize(
    context: MlonMcuContext, params=None, rebuild=False, verbose=False, threads=multiprocessing.cpu_count()
):
    """Download the visualize.py script for TFLite."""
    # This script is content of the tensorflow repo (not tflite-micro) and unfortunately not bundled
    # into the tensorflow python package. Therefore just download this single file form GitHub

    tfLiteVizualizeName = utils.makeDirName("tflite_visualize")
    tfLiteVizualizeInstallDir = context.environment.paths["deps"].path / "install" / tfLiteVizualizeName
    tfLiteVizualizeExe = tfLiteVizualizeInstallDir / "visualize.py"
    user_vars = context.environment.vars
    if "tflite_visualize.exe" in user_vars:  # TODO: also check command line flags?
        return False
    if rebuild or not utils.is_populated(tfLiteVizualizeInstallDir):
        tfLiteVizualizeInstallDir.mkdir(exist_ok=True)
        url = "https://raw.githubusercontent.com/tensorflow/tensorflow/master/tensorflow/lite/tools/visualize.py"
        utils.download(url, tfLiteVizualizeExe, progress=verbose)
    context.cache["tflite_visualize.exe"] = tfLiteVizualizeExe
    context.export_paths.add(tfLiteVizualizeInstallDir)



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MLonMCU RISC-V Target definitions"""

from pathlib import Path

from mlonmcu.logging import get_logger
from mlonmcu.feature.features import SUPPORTED_TVM_BACKENDS
from mlonmcu.target import Target
from mlonmcu.config import str2list, str2bool
from .util import sort_extensions_canonical, join_extensions, update_extensions, split_extensions

logger = get_logger()


class RISCVTarget(Target):
    """Common base class for RISCV-like targets. Please do not use this as a target itself!"""

    FEATURES = Target.FEATURES

    DEFAULTS = {
        **Target.DEFAULTS,
        # Default: rv32gc
        "xlen": 32,
        "embedded": False,
        "compressed": True,
        "atomic": True,
        "multiply": True,
        "extra_args": "",
        "timeout_sec": 0,  # disabled
        "extensions": [],  # Should only be used for unhandled custom exts
        "fpu": "double",  # allowed: none, single, double
        "arch": None,  # Please use above properties if possible
        "abi": None,  # Please use above properties if possible
        "attr": "",  # Please avoid using this directly
        "cpu": None,
    }
    REQUIRED = {"riscv_gcc.install_dir", "riscv_gcc.name", "riscv_gcc.variant"}
    PUPL_GCC_TOOLCHAIN_REQUIRED = {"pulp_gcc.install_dir", "pulp_gcc.name"}  # TODO elegant handle customized toolchain
    OPTIONAL = {"llvm.install_dir", "mlif.toolchain"}  # TODO: just a workaround until tc components are implemented

    def reconfigure(self):
        # super().reconfigure()
        self.config.update(
            {
                "final_arch": self.arch,
                "final_abi": self.abi,
            }
        )

    @property
    def riscv_gcc_prefix(self):
        return Path(self.config["riscv_gcc.install_dir"])

    @property
    def riscv_gcc_basename(self):
        return Path(self.config["riscv_gcc.name"])

    @property
    def pulp_gcc_prefix(self):
        return Path(self.config["pulp_gcc.install_dir"])

    @property
    def pulp_gcc_basename(self):
        return Path(self.config["pulp_gcc.name"])

    @property
    def gcc_variant(self):
        return self.config["riscv_gcc.variant"]

    @property
    def xlen(self):
        arch = self.config.get("arch", None)
        if arch is not None:
            assert isinstance(arch, str)
            xlen = int(arch[2:4])
            return xlen
        return int(self.config["xlen"])

    @property
    def embedded(self):
        arch = self.config.get("arch", None)
        if arch is not None:
            exts = split_extensions(arch)
            if "e" in exts:
                return True
        value = self.config["embedded"]
        return str2bool(value)

    @property
    def compressed(self):
        arch = self.config.get("arch", None)
        if arch is not None:
            exts = split_extensions(arch)
            if "c" in exts:
                return True
        value = self.config["compressed"]
        return str2bool(value)

    @property
    def atomic(self):
        arch = self.config.get("arch", None)
        if arch is not None:
            exts = split_extensions(arch)
            if "a" in exts or "g" in exts:
                return True
        value = self.config["atomic"]
        return str2bool(value)

    @property
    def multiply(self):
        arch = self.config.get("arch", None)
        if arch is not None:
            exts = split_extensions(arch)
            if "m" in exts or "g" in exts:
                return True
        value = self.config["multiply"]
        return str2bool(value)

    @property
    def extensions(self):
        arch = self.config.get("arch", None)
        exts = self.config.get("extensions", []).copy()
        if isinstance(exts, str):
            exts = str2list(exts)
        assert isinstance(exts, (list, set))
        exts = set(exts)

        if arch is not None:
            assert isinstance(arch, str)
            exts = split_extensions(arch) | exts
        return update_extensions(
            exts,
            fpu=self.fpu,
            embedded=self.embedded,
            compressed=self.compressed,
            atomic=self.atomic,
            multiply=self.multiply,
            pext=False,
            pext_spec=None,
            vext=False,
            elen=None,
            embedded_vext=False,
            vlen=None,
        )

    @property
    def gcc_extensions(self):
        # return [ext for ext in (self.extensions | {"zicsr"}) if ext not in ["xcorev", "xcorevmac", "xcorevmem"]]
        exts = {"zicsr"}
        for ext in self.extensions:
            if "xcv" in ext:
                if ext[-2] != "p":
                    # Add missing version
                    ext = ext + "1p0"
            exts.add(ext)
        return exts

    @property
    def llvm_extensions(self):
        return [ext for ext in self.extensions if ext not in ["zifencei", "zicsr"]]

    @property
    def arch(self):
        temp = self.config["arch"]  # TODO: allow underscores and versions
        if temp:
            return temp
        else:
            exts_str = join_extensions(sort_extensions_canonical(self.extensions, lower=True))
            return f"rv{self.xlen}{exts_str}"

    @property
    def cpu(self):
        temp = self.config["cpu"]
        if temp:
            return temp
        else:
            return f"generic-rv{self.xlen}"

    @property
    def llvm_arch(self):
        temp = self.config["arch"]  # TODO: allow underscores and versions
        if temp:
            return temp
        else:
            exts_str = join_extensions(sort_extensions_canonical(self.llvm_extensions, lower=True))
            return f"rv{self.xlen}{exts_str}"

    @property
    def gcc_arch(self):
        temp = self.config["arch"]  # TODO: allow underscores and versions
        if temp:
            return temp
        else:
            exts_str = join_extensions(sort_extensions_canonical(self.gcc_extensions, lower=True))
            return f"rv{self.xlen}{exts_str}"

    @property
    def abi(self):
        temp = self.config["abi"]
        if temp:
            return temp
        else:
            if self.xlen == 32:
                temp = "ilp32"
            elif self.xlen == 64:
                temp = "lp64"
            else:
                raise RuntimeError(f"Invalid xlen: {self.xlen}")
            if "d" in self.extensions or "g" in self.extensions:
                temp += "d"
            elif "f" in self.extensions:
                temp += "f"
            return temp

    @property
    def attrs(self):
        attr = self.config.get("attr", None)
        attrs = []
        if attr is not None:
            assert isinstance(attr, str)
            attrs = attr.split(",")
            if len(attrs) == 1 and len(attrs[0]) == 0:
                attrs = []
        if len(attrs) == 0:
            for ext in sort_extensions_canonical(self.extensions, lower=True, unpack=True):
                if ext == "i":
                    continue
                attrs.append(f"+{ext}")
            if self.xlen == 64:
                attrs.append("+64bit")
        return attrs

    @property
    def attr(self):
        attrs = self.attrs
        return ",".join(attrs)

    @property
    def extra_args(self):
        ret = self.config["extra_args"]
        if isinstance(ret, str):
            if len(ret) == 0:
                ret = []
            else:
                ret = [ret]  # TODO: properly split quoted args
        return ret

    @property
    def timeout_sec(self):
        return float(self.config["timeout_sec"])

    @property
    def fpu(self):
        arch = self.config.get("arch", None)
        if arch is not None:
            exts = split_extensions(arch)
            if "d" in exts or "g" in exts:
                return "double"
            elif "f" in exts:
                return "single"
            else:
                return "none"
        value = self.config["fpu"]
        if value is None or not value:
            value = "none"
        assert value in ["none", "single", "double"]
        return value

    @property
    def has_fpu(self):
        return self.fpu != "none"

    @property
    def toolchain(self):
        value = self.config.get("mlif.toolchain", None)
        if value is None:
            value = "gcc"
        return str(value)

    def get_target_system(self):
        return "generic_riscv"  # TODO: rename to generic-rv32 for compatibility with LLVM

    def get_platform_defs(self, platform):
        ret = super().get_platform_defs(platform)
        # TODO refactor the following using inheritance instead of branching
        if "riscv_gcc.install_dir" in self.REQUIRED:  # the target chooses to use the riscv_gcc toolchain
            ret["RISCV_ELF_GCC_PREFIX"] = self.riscv_gcc_prefix
            ret["RISCV_ELF_GCC_BASENAME"] = self.riscv_gcc_basename
        elif "pulp_gcc.install_dir" in self.REQUIRED:  # the target chooses to use the pulp_gcc toolchain
            ret["RISCV_ELF_GCC_PREFIX"] = self.pulp_gcc_prefix
            ret["RISCV_ELF_GCC_BASENAME"] = self.pulp_gcc_basename
        ret["RISCV_ARCH"] = self.gcc_arch if self.toolchain == "gcc" else self.llvm_arch
        ret["RISCV_ABI"] = self.abi
        ret["RISCV_MCPU"] = self.cpu
        # llvm/clang only!
        ret["RISCV_ATTR"] = self.attr

        def feature_helper(attrs):
            # TODO
            return ""

        ret["RISCV_FEATURES"] = feature_helper(self.attr)
        return ret

    def get_arch(self):
        return "riscv"

    def get_backend_config(self, backend, optimized_layouts=False, optimized_schedules=False):
        ret = {}
        if backend in SUPPORTED_TVM_BACKENDS:
            arch_clean = self.llvm_arch.replace("imafd", "g").replace("_", "-")
            ret.update(
                {
                    # "target_march": self.llvm_arch,
                    "target_mtriple": self.riscv_gcc_basename,
                    "target_mabi": self.abi,
                    "target_mattr": self.attr,
                    "target_mcpu": self.cpu,
                    "target_model": f"etiss-{arch_clean}",
                    # "target_model": f"{self.name}-{arch_clean}",
                    "target_num_cores": 1,  # TODO: also add for non-riscv targets
                    # "target_device": ?,
                    # "target_libs": ?,
                    # "target_tag": ?,
                    # "target_march": ?,
                    # "target_keys": ?,
                    # "target_opt_level": ?,
                    # "target_cl_opt": ?,
                    # "target_mfloat_abi": ?,
                    # "target_fast_math_ninf": ?,
                    # "target_fast_math_contract": ?,
                    # "target_fast_math_nnan": ?,
                    # "target_fast_math": ?,
                    # "target_fast_math_nsz": ?,
                    # "target_fast_math_reassoc": ?,
                    # "target_fast_math_arcp": ?,
                    # "target_model": "host",
                }
            )
            if optimized_schedules:
                ret.update(
                    {
                        "target_device": "riscv_cpu",
                        "target_keys": None,
                    }
                )
        return ret



#
# Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
#
# This file is part of MLonMCU.
# See https://github.com/tum-ei-eda/mlonmcu.git for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""TVM RPC Platform"""
from mlonmcu.config import str2bool
from ...platform import Platform


class TvmRpcPlatform(Platform):
    """TVM RPC platform class."""

    FEATURES = Platform.FEATURES | {
        "tvm_rpc",
    }

    DEFAULTS = {
        **Platform.DEFAULTS,
        "use_rpc": False,
        "rpc_key": None,
        "rpc_hostname": None,
        "rpc_port": None,
    }

    @property
    def use_rpc(self):
        value = self.config["use_rpc"]
        return str2bool(value) if not isinstance(value, (bool, int)) else value

    @property
    def rpc_key(self):
        return self.config["rpc_key"]

    @property
    def rpc_hostname(self):
        return self.config["rpc_hostname"]

    @property
    def rpc_port(self):
        return self.config["rpc_port"]
